from typing import Any, Optional
import time
import re

from netmiko.base_connection import BaseConnection


class DellIsilonSSH(BaseConnection):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self.ansi_escape_codes = True
        self._test_channel_read(pattern=r"[#\$]")
        self._zsh_mode()
        self.find_prompt()
        self.set_base_prompt()

    def set_base_prompt(
        self,
        pri_prompt_terminator: str = "$",
        alt_prompt_terminator: str = "#",
        delay_factor: float = 1.0,
        pattern: Optional[str] = None,
    ) -> str:
        """Determine base prompt."""
        return super().set_base_prompt(
            pri_prompt_terminator=pri_prompt_terminator,
            alt_prompt_terminator=alt_prompt_terminator,
            delay_factor=delay_factor,
            pattern=pattern,
        )

    def strip_ansi_escape_codes(self, string_buffer: str) -> str:
        """Remove Null code"""
        output = re.sub(r"\x00", "", string_buffer)
        return super().strip_ansi_escape_codes(output)

    def _zsh_mode(self, prompt_terminator: str = "$") -> None:
        """Run zsh command to unify the environment"""
        if self.global_delay_factor < 1:
            delay_factor = 1.0
        else:
            delay_factor = self.global_delay_factor
        command = self.RETURN + "zsh" + self.RETURN
        self.write_channel(command)
        time.sleep(0.25 * delay_factor)
        self._set_prompt(prompt_terminator)
        time.sleep(0.25 * delay_factor)
        self.clear_buffer()

    def _set_prompt(self, prompt_terminator: str = "$") -> None:
        prompt = f"PROMPT='%m{prompt_terminator}'"
        command = self.RETURN + prompt + self.RETURN
        self.write_channel(command)

    def disable_paging(self, *args: Any, **kwargs: Any) -> str:
        """Isilon doesn't have paging by default."""
        return ""

    def check_enable_mode(self, check_string: str = "#") -> bool:
        return super().check_enable_mode(check_string=check_string)

    def enable(
        self,
        cmd: str = "sudo su",
        pattern: str = "ssword",
        enable_pattern: Optional[str] = None,
        re_flags: int = re.IGNORECASE,
    ) -> str:
        delay_factor = self.select_delay_factor(delay_factor=1)
        output = ""
        if not self.check_enable_mode():
            output += self._send_command_timing_str(
                cmd, strip_prompt=False, strip_command=False
            )
            if re.search(pattern, output, flags=re_flags):
                self.write_channel(self.normalize_cmd(self.secret))
            output += self.read_until_pattern(pattern=r"#.*$")
            time.sleep(1 * delay_factor)
            self._set_prompt(prompt_terminator="#")
            if not self.check_enable_mode():
                raise ValueError("Failed to enter enable mode")
        return output

    def exit_enable_mode(self, exit_command: str = "exit") -> str:
        return super().exit_enable_mode(exit_command=exit_command)

    def check_config_mode(self, check_string: str = "#", pattern: str = "") -> bool:
        """Use equivalent enable method."""
        return self.check_enable_mode(check_string=check_string)

    def config_mode(
        self,
        config_command: str = "sudo su",
        pattern: str = "ssword",
        re_flags: int = re.IGNORECASE,
    ) -> str:
        """Use equivalent enable method."""
        return self.enable(cmd=config_command, pattern=pattern, re_flags=re_flags)

    def exit_config_mode(self, exit_config: str = "exit", pattern: str = "") -> str:
        """Use equivalent enable method."""
        return self.exit_enable_mode(exit_command=exit_config)



from datetime import datetime
import pytest
import time
import re
from netmiko import ReadTimeout


def execute_cmd(conn, pattern, read_timeout, cmd="show tech-support\n", max_loops=None):
    conn.write_channel("show tech-support\n")
    return conn.read_until_pattern(
        pattern=pattern, read_timeout=read_timeout, max_loops=max_loops
    )


def my_cleanup(conn, sleep=180):
    try:
        # Must be long enough for "show tech-support" to finish
        time.sleep(sleep)
        conn.disconnect()
    except Exception:
        # If it fails, just try to keep going
        pass


def show_long_running(conn, read_timeout, expect_string):
    start_time = datetime.now()
    my_exception = None
    try:
        execute_cmd(conn, pattern=expect_string, read_timeout=read_timeout)
    except Exception as e:
        my_exception = e
    finally:
        end_time = datetime.now()
        exec_time = end_time - start_time
        # clean-up
        my_cleanup(conn)
        return (my_exception, exec_time)


def show_long_running_notimeout(conn, read_timeout):
    """This execution should work i.e. no exception."""
    start_time = datetime.now()

    my_prompt = conn.find_prompt()
    pattern = re.escape(my_prompt)

    output = execute_cmd(conn, pattern=pattern, read_timeout=read_timeout)
    end_time = datetime.now()
    exec_time = end_time - start_time
    return (output, exec_time)


def test_read_longrunning_cmd(net_connect_newconn):

    read_timeout = 300
    output, exec_time = show_long_running_notimeout(
        net_connect_newconn, read_timeout=read_timeout
    )
    assert "show interface" in output
    assert exec_time.total_seconds() > 10


@pytest.mark.parametrize(
    "test_timeout,allowed_percentage",
    [(0.4, 5.0), (1, 2.0), (5, 1.0), (10, 0.5), (60, 0.2)],
)
def test_read_timeout(net_connect_newconn, test_timeout, allowed_percentage):

    # Explicitly send expect_string so timing is more accurate
    my_prompt = net_connect_newconn.find_prompt()
    pattern = re.escape(my_prompt)

    my_except, exec_time = show_long_running(
        net_connect_newconn, read_timeout=test_timeout, expect_string=pattern
    )

    # Returned exception should be read_timeout
    assert isinstance(my_except, ReadTimeout)

    # Calculate difference in execution time from timeout
    time_diff = abs(exec_time.total_seconds() - test_timeout)
    # Convert difference to a percentage of expected timeout
    time_margin_percent = time_diff / test_timeout * 100
    # Margin off should be less than the allowed_percentage
    assert time_margin_percent < allowed_percentage


@pytest.mark.parametrize(
    "test_timeout,allowed_percentage", [(1, 2.0), (5, 1.0), (10, 0.5), (60, 0.2)]
)
def test_read_timeout_override(net_connect_newconn, test_timeout, allowed_percentage):

    net_connect_newconn.read_timeout_override = 12
    ssh_conn = net_connect_newconn

    # Explicitly send expect_string so timing is more accurate
    my_prompt = net_connect_newconn.find_prompt()
    pattern = re.escape(my_prompt)

    my_except, exec_time = show_long_running(
        ssh_conn, read_timeout=test_timeout, expect_string=pattern
    )

    # Returned exception should be read_timeout
    assert isinstance(my_except, ReadTimeout)

    # For comparsions compare to the override time with a fixed allowed percentage
    test_timeout = ssh_conn.read_timeout_override
    allowed_percentage = 0.5

    # Calculate difference in execution time from read_timeout_override
    time_diff = abs(exec_time.total_seconds() - test_timeout)

    # Convert difference to a percentage of expected timeout
    time_margin_percent = time_diff / test_timeout * 100
    # Margin off should be less than the allowed_percentage
    assert time_margin_percent < allowed_percentage


def test_deprecation_max_loops(net_connect_newconn):

    nc = net_connect_newconn
    cmd = "show ip int brief"
    with pytest.deprecated_call():
        execute_cmd(nc, pattern=r"#", read_timeout=10, cmd=cmd, max_loops=1000)



from typing import Optional
from netmiko.cisco_base_connection import CiscoSSHConnection
import re


class EndaceSSH(CiscoSSHConnection):
    def session_preparation(self) -> None:
        self._test_channel_read()
        self.set_base_prompt()
        self.set_terminal_width()
        self.disable_paging(command="no cli session paging enable")

    def enable(
        self,
        cmd: str = "enable",
        pattern: str = "",
        enable_pattern: Optional[str] = None,
        re_flags: int = re.IGNORECASE,
    ) -> str:
        return super().enable(
            cmd=cmd, pattern=pattern, enable_pattern=enable_pattern, re_flags=re_flags
        )

    def check_config_mode(
        self, check_string: str = "(config) #", pattern: str = ""
    ) -> bool:
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def config_mode(
        self, config_command: str = "conf t", pattern: str = "", re_flags: int = 0
    ) -> str:
        output = ""
        if not self.check_config_mode():
            output += self._send_command_timing_str(
                config_command, strip_command=False, strip_prompt=False
            )
            if "to enter configuration mode anyway" in output:
                output += self._send_command_timing_str(
                    "YES", strip_command=False, strip_prompt=False
                )
            if not self.check_config_mode():
                raise ValueError("Failed to enter configuration mode")
        return output

    def exit_config_mode(self, exit_config: str = "exit", pattern: str = "#") -> str:
        return super().exit_config_mode(exit_config=exit_config, pattern=pattern)

    def save_config(
        self,
        cmd: str = "configuration write",
        confirm: bool = False,
        confirm_response: str = "",
    ) -> str:
        self.enable()
        self.config_mode()
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )



from typing import Optional, Any
import time
import warnings
from netmiko.no_enable import NoEnable
from netmiko.base_connection import DELAY_FACTOR_DEPR_SIMPLE_MSG
from netmiko.cisco_base_connection import CiscoBaseConnection


class SixwindOSBase(NoEnable, CiscoBaseConnection):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self.ansi_escape_codes = True
        self._test_channel_read()
        self.set_base_prompt()
        # Clear the read buffer
        time.sleep(0.3 * self.global_delay_factor)
        self.clear_buffer()

    def disable_paging(self, *args: Any, **kwargs: Any) -> str:
        """6WIND requires no-pager at the end of command, not implemented at this time."""
        pass

    def set_base_prompt(
        self,
        pri_prompt_terminator: str = ">",
        alt_prompt_terminator: str = "#",
        delay_factor: float = 1.0,
        pattern: Optional[str] = None,
    ) -> str:
        """Sets self.base_prompt: used as delimiter for stripping of trailing prompt in output."""

        prompt = super().set_base_prompt(
            pri_prompt_terminator=pri_prompt_terminator,
            alt_prompt_terminator=alt_prompt_terminator,
            delay_factor=delay_factor,
            pattern=pattern,
        )
        prompt = prompt.strip()
        self.base_prompt = prompt
        return self.base_prompt

    def config_mode(
        self, config_command: str = "edit running", pattern: str = "", re_flags: int = 0
    ) -> str:
        return super().config_mode(
            config_command=config_command, pattern=pattern, re_flags=re_flags
        )

    def commit(
        self,
        comment: str = "",
        read_timeout: float = 120.0,
        delay_factor: Optional[float] = None,
    ) -> str:
        """
        Commit the candidate configuration.

        Raise an error and return the failure if the commit fails.

        delay_factor: Deprecated in Netmiko 4.x. Will be eliminated in Netmiko 5.

        """

        if delay_factor is not None:
            warnings.warn(DELAY_FACTOR_DEPR_SIMPLE_MSG, DeprecationWarning)

        error_marker = "Failed to generate committed config"
        command_string = "commit"

        output = self.config_mode()
        output += self._send_command_str(
            command_string,
            strip_prompt=False,
            strip_command=False,
            read_timeout=read_timeout,
            expect_string=r"#",
        )
        output += self.exit_config_mode()

        if error_marker in output:
            raise ValueError(f"Commit failed with following errors:\n\n{output}")

        return output

    def exit_config_mode(self, exit_config: str = "exit", pattern: str = r">") -> str:
        """Exit configuration mode."""

        return super().exit_config_mode(exit_config=exit_config, pattern=pattern)

    def check_config_mode(self, check_string: str = "#", pattern: str = "") -> bool:
        """Checks whether in configuration mode. Returns a boolean."""

        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def save_config(
        self,
        cmd: str = "copy running startup",
        confirm: bool = True,
        confirm_response: str = "y",
    ) -> str:
        """Save Config for 6WIND"""

        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )


class SixwindOSSSH(SixwindOSBase):

    pass



from typing import Any, Union, Sequence, TextIO, Optional
from netmiko.no_config import NoConfig
from netmiko.cisco_base_connection import CiscoSSHConnection


class CloudGenixIonSSH(NoConfig, CiscoSSHConnection):
    def establish_connection(self, width: int = 100, height: int = 1000) -> None:
        super().establish_connection(width=width, height=height)

    def session_preparation(self, *args: Any, **kwargs: Any) -> None:
        """Prepare the session after the connection has been established."""
        self.ansi_escape_codes = True
        self._test_channel_read(pattern=r"[>#]")
        self.write_channel(self.RETURN)
        self.set_base_prompt(delay_factor=5)

    def disable_paging(self, *args: Any, **kwargs: Any) -> str:
        """Cloud Genix ION sets terminal height in establish_connection"""
        return ""

    def find_prompt(
        self, delay_factor: float = 1.0, pattern: Optional[str] = None
    ) -> str:
        prompt = super().find_prompt(delay_factor=delay_factor, pattern=pattern)
        prompt = self.strip_backspaces(prompt).strip()
        return prompt

    def strip_command(self, command_string: str, output: str) -> str:
        output = super().strip_command(command_string, output)
        # command_string gets repainted potentially multiple times (grab everything after last one)
        output = output.split(command_string)[-1]
        return output

    def save_config(self, *args: Any, **kwargs: Any) -> str:
        """Not Implemented"""
        raise NotImplementedError

    def send_config_set(
        self,
        config_commands: Union[str, Sequence[str], TextIO, None] = None,
        exit_config_mode: bool = False,
        **kwargs: Any
    ) -> str:
        return super().send_config_set(
            config_commands=config_commands, exit_config_mode=exit_config_mode, **kwargs
        )



#!/usr/bin/env python
from netmiko.ssh_autodetect import SSH_MAPPER_BASE


def test_ssh_base_mapper_order():
    "SSH_MAPPER_BASE should be sorted based on the most common command used." ""
    assert SSH_MAPPER_BASE[0][1]["cmd"] == "show version"



#!/usr/bin/env python
import pytest
import logging
from netmiko import ConnectHandler, ConnLogOnly, ConnUnify
from netmiko import NetmikoAuthenticationException
from netmiko import NetmikoTimeoutException
from netmiko import ConnectionException


# Fictional devices that will fail
DEVICE1 = {
    "host": "localhost",
    "device_type": "linux",
    "username": "bogus",
    "password": "bogus",
}

DEVICE2 = {
    "host": "localhost",
    "device_type": "linux",
    "username": "bogus",
    "password": "bogus",
    "port": 8022,
}


def test_connecthandler():
    with pytest.raises(NetmikoAuthenticationException):
        net_connect = ConnectHandler(**DEVICE1)
    with pytest.raises(NetmikoTimeoutException):
        net_connect = ConnectHandler(**DEVICE2)  # noqa


def test_connlogonly(caplog):
    log_level = logging.INFO
    log_file = "my_output.log"

    net_connect = ConnLogOnly(log_file=log_file, log_level=log_level, **DEVICE1)
    assert net_connect is None
    net_connect = ConnLogOnly(log_file=log_file, log_level=log_level, **DEVICE2)
    assert net_connect is None

    assert "ERROR" in caplog.text
    assert caplog.text.count("ERROR") == 2
    assert "Authentication failure" in caplog.text
    assert "was unable to reach the provided host and port" in caplog.text


def test_connunify():
    with pytest.raises(ConnectionException):
        net_connect = ConnUnify(**DEVICE1)
    with pytest.raises(ConnectionException):
        net_connect = ConnUnify(**DEVICE2)  # noqa



"""Subclass specific to Cisco Viptela."""
from typing import Union, Sequence, TextIO, Any
import re

from netmiko.cisco_base_connection import CiscoSSHConnection


class CiscoViptelaSSH(CiscoSSHConnection):
    """Subclass specific to Cisco Viptela."""

    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self.disable_paging(command="paginate false")

    def check_config_mode(self, check_string: str = ")#", pattern: str = "#") -> bool:
        """Checks if the device is in configuration mode or not."""
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def commit(self, confirm: bool = False, confirm_response: str = "") -> str:
        cmd = "commit"
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )

    def config_mode(
        self,
        config_command: str = "conf terminal",
        pattern: str = "",
        re_flags: int = 0,
    ) -> str:
        return super().config_mode(
            config_command=config_command, pattern=pattern, re_flags=re_flags
        )

    def send_config_set(
        self,
        config_commands: Union[str, Sequence[str], TextIO, None] = None,
        exit_config_mode: bool = False,
        **kwargs: Any,
    ) -> str:
        return super().send_config_set(
            config_commands=config_commands, exit_config_mode=exit_config_mode, **kwargs
        )

    def exit_config_mode(self, exit_config: str = "end", pattern: str = r"#") -> str:
        """
        Exit from configuration mode.

        Viptela might have the following in the output (if no 'commit()' occurred.

        Uncommitted changes found, commit them? [yes/no/CANCEL]
        """
        output = ""
        if self.check_config_mode():
            self.write_channel(self.normalize_cmd(exit_config))
            # Make sure you read until you detect the command echo (avoid getting out of sync)
            if self.global_cmd_verify is not False:
                output += self.read_until_pattern(
                    pattern=re.escape(exit_config.strip())
                )
            if not re.search(pattern, output, flags=re.M):
                uncommit_pattern = r"Uncommitted changes found"
                new_pattern = f"({pattern}|{uncommit_pattern})"
                output += self.read_until_pattern(pattern=new_pattern)
                # Do not save 'uncommited changes'
                if uncommit_pattern in output:
                    self.write_channel(self.normalize_cmd("no"))
                    output += self.read_until_pattern(pattern=pattern)

            if self.check_config_mode():
                raise ValueError("Failed to exit configuration mode")
        return output

    def save_config(
        self, cmd: str = "commit", confirm: bool = False, confirm_response: str = ""
    ) -> str:
        """Saves Config"""
        raise NotImplementedError



from typing import Optional
from netmiko.no_config import NoConfig
from netmiko.base_connection import BaseConnection


class NetscalerSSH(NoConfig, BaseConnection):
    """Netscaler SSH class."""

    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        cmd = f"{self.RETURN}set cli mode -page OFF{self.RETURN}"
        self.disable_paging(command=cmd)
        self.set_base_prompt()

    def set_base_prompt(
        self,
        pri_prompt_terminator: str = ">",
        alt_prompt_terminator: str = "#",
        delay_factor: float = 1.0,
        pattern: Optional[str] = None,
    ) -> str:
        """Sets self.base_prompt.

        Netscaler has only '>' for the prompt.
        """
        base_prompt = super().set_base_prompt(
            pri_prompt_terminator=pri_prompt_terminator,
            alt_prompt_terminator=alt_prompt_terminator,
            delay_factor=delay_factor,
            pattern=pattern,
        )
        # If null-string, set base_prompt to just ">"
        if not base_prompt:
            self.base_prompt = pri_prompt_terminator
        return self.base_prompt

    def strip_prompt(self, a_string: str) -> str:
        """Strip 'Done' from command output"""
        output = super().strip_prompt(a_string)
        lines = output.split(self.RESPONSE_RETURN)
        if "Done" in lines[-1]:
            return self.RESPONSE_RETURN.join(lines[:-1])
        else:
            return output



from typing import Any, Optional
import time

from netmiko.no_enable import NoEnable
from netmiko.no_config import NoConfig
from netmiko.cisco.cisco_ios import CiscoIosBase


class KeymileSSH(NoEnable, NoConfig, CiscoIosBase):
    def __init__(self, **kwargs: Any) -> None:
        kwargs.setdefault("default_enter", "\r\n")
        super().__init__(**kwargs)

    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read(pattern=r">")
        self.set_base_prompt()
        time.sleep(0.3 * self.global_delay_factor)
        self.clear_buffer()

    def disable_paging(self, *args: Any, **kwargs: Any) -> str:
        """Keymile does not use paging."""
        return ""

    def strip_prompt(self, a_string: str) -> str:
        """Remove appending empty line and prompt from output"""
        a_string = a_string[:-1]
        return super().strip_prompt(a_string=a_string)

    def set_base_prompt(
        self,
        pri_prompt_terminator: str = ">",
        alt_prompt_terminator: str = ">",
        delay_factor: float = 1.0,
        pattern: Optional[str] = None,
    ) -> str:
        """set prompt termination to >"""
        return super().set_base_prompt(
            pri_prompt_terminator=pri_prompt_terminator,
            alt_prompt_terminator=alt_prompt_terminator,
            delay_factor=delay_factor,
            pattern=pattern,
        )



from typing import Any

from netmiko.no_enable import NoEnable
from netmiko.base_connection import BaseConnection


class NetAppcDotSSH(NoEnable, BaseConnection):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self.set_base_prompt()
        cmd = self.RETURN + "rows 0" + self.RETURN
        self.disable_paging(command=cmd)

    def send_command_with_y(self, *args: Any, **kwargs: Any) -> str:
        output = self._send_command_timing_str(*args, **kwargs)
        if "{y|n}" in output:
            output += self._send_command_timing_str(
                "y", strip_prompt=False, strip_command=False
            )
        return output

    def check_config_mode(self, check_string: str = "*>", pattern: str = "") -> bool:
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def config_mode(
        self,
        config_command: str = "set -privilege diagnostic -confirmations off",
        pattern: str = "",
        re_flags: int = 0,
    ) -> str:
        return super().config_mode(
            config_command=config_command, pattern=pattern, re_flags=re_flags
        )

    def exit_config_mode(
        self,
        exit_config: str = "set -privilege admin -confirmations off",
        pattern: str = "",
    ) -> str:
        return super().exit_config_mode(exit_config=exit_config, pattern=pattern)



"""
Aruba OS support.

For use with Aruba OS Controllers.

"""
from typing import Any
from netmiko.cisco_base_connection import CiscoSSHConnection


class ArubaSSH(CiscoSSHConnection):
    """Aruba OS support"""

    def __init__(self, **kwargs: Any) -> None:
        if kwargs.get("default_enter") is None:
            kwargs["default_enter"] = "\r"
        # Aruba has an auto-complete on space behavior that is problematic
        if kwargs.get("global_cmd_verify") is None:
            kwargs["global_cmd_verify"] = False
        return super().__init__(**kwargs)

    def session_preparation(self) -> None:
        """Aruba OS requires enable mode to disable paging."""
        # Aruba switches output ansi codes
        self.ansi_escape_codes = True
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self.enable()
        self.disable_paging(command="no page")

    def check_config_mode(
        self, check_string: str = "(config) #", pattern: str = r"[>#]"
    ) -> bool:
        """
        Checks if the device is in configuration mode or not.

        Aruba uses "(<controller name>) (config) #" as config prompt
        """
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def config_mode(
        self,
        config_command: str = "configure term",
        pattern: str = "",
        re_flags: int = 0,
    ) -> str:
        """Aruba auto completes on space so 'configure' needs fully spelled-out."""
        return super().config_mode(
            config_command=config_command, pattern=pattern, re_flags=re_flags
        )



from netmiko.no_config import NoConfig
from netmiko.base_connection import BaseConnection


class CheckPointGaiaSSH(NoConfig, BaseConnection):
    """
    Implements methods for communicating with Check Point Gaia
    firewalls.
    """

    def session_preparation(self) -> None:
        """
        Prepare the session after the connection has been established.

        Set the base prompt for interaction ('>').
        """
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self.disable_paging(command="set clienv rows 0")

    def save_config(
        self, cmd: str = "", confirm: bool = False, confirm_response: str = ""
    ) -> str:
        raise NotImplementedError



from netmiko.no_config import NoConfig
from netmiko.base_connection import BaseConnection


class F5TmshSSH(NoConfig, BaseConnection):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read(pattern=r"#")
        self.tmsh_mode()
        self._config_mode = False
        cmd = 'run /util bash -c "stty cols 255"'
        self.set_terminal_width(command=cmd, pattern="run")
        self.disable_paging(
            command="modify cli preference pager disabled display-threshold 0"
        )

    def tmsh_mode(self, delay_factor: float = 1.0) -> str:
        """tmsh command is equivalent to config command on F5."""
        command = f"{self.RETURN}tmsh{self.RETURN}"
        output = self._send_command_str(command, expect_string=r"tmos.*#")
        self.set_base_prompt()
        return output

    def exit_tmsh(self) -> str:
        output = self._send_command_str("quit", expect_string=r"#")
        self.set_base_prompt()
        return output

    def cleanup(self, command: str = "exit") -> None:
        """Gracefully exit the SSH session."""
        try:
            self.exit_tmsh()
        except Exception:
            pass

        # Always try to send final 'exit' (command)
        self._session_log_fin = True
        self.write_channel(command + self.RETURN)



"""Support for Extreme TierraOS."""
import time
from netmiko.no_enable import NoEnable
from netmiko.cisco_base_connection import CiscoSSHConnection


class ExtremeTierraSSH(NoEnable, CiscoSSHConnection):
    """Support for Extreme TierraOS."""

    def session_preparation(self) -> None:
        self._test_channel_read(pattern=r"#")
        self.set_base_prompt()
        self.set_terminal_width()
        self.disable_paging(command="terminal length 0")

    def special_login_handler(self, delay_factor: float = 1.0) -> None:
        """Adding a delay after login."""
        delay_factor = self.select_delay_factor(delay_factor)
        self.write_channel(self.RETURN)
        time.sleep(1 * delay_factor)

    def save_config(
        self,
        cmd: str = "copy run flash://config-file/startup-config",
        confirm: bool = False,
        confirm_response: str = "y",
    ) -> str:
        """Save Config for Extreme TierraOS."""
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )



import time
from netmiko.vyos.vyos_ssh import VyOSSSH


class UbiquitiEdgeRouterSSH(VyOSSSH):
    """Implement methods for interacting with EdgeOS EdgeRouter network devices."""

    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read()
        self.set_base_prompt()
        self.set_terminal_width(command="terminal width 512")
        self.disable_paging(command="terminal length 0")
        # Clear the read buffer
        time.sleep(0.3 * self.global_delay_factor)
        self.clear_buffer()

    def save_config(
        self, cmd: str = "save", confirm: bool = False, confirm_response: str = ""
    ) -> str:
        """Saves Config."""
        if confirm is True:
            raise ValueError("EdgeRouter does not support save_config confirmation.")
        output = self._send_command_str(command_string=cmd)
        if "Done" not in output:
            raise ValueError(f"Save failed with following errors:\n\n{output}")
        return output



from netmiko.cisco_base_connection import CiscoSSHConnection


class CiscoS300Base(CiscoSSHConnection):
    """
    Support for Cisco SG300 series of devices.

    Note, must configure the following to disable SG300 from prompting for username twice:

    configure terminal
    ip ssh password-auth
    """

    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self.ansi_escape_codes = True
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self.set_terminal_width(command="terminal width 511", pattern="terminal")
        self.disable_paging(command="terminal datadump")

    def save_config(
        self,
        cmd: str = "write memory",
        confirm: bool = True,
        confirm_response: str = "Y",
    ) -> str:
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )


class CiscoS300SSH(CiscoS300Base):
    pass


class CiscoS300Telnet(CiscoS300Base):
    """
    Support for Cisco SG300 series of devices, with telnet.
    Note: can be used with Sx200 series, with telnet enabled.
    """

    pass



"""Extreme Virtual Services Platform Support."""
from netmiko.cisco_base_connection import CiscoSSHConnection


class ExtremeVspSSH(CiscoSSHConnection):
    """Extreme Virtual Services Platform Support."""

    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self.disable_paging(command="terminal more disable")

    def save_config(
        self,
        cmd: str = "save config",
        confirm: bool = False,
        confirm_response: str = "",
    ) -> str:
        """Save Config"""
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )



from typing import Optional
import re


class NoEnable:
    """
    Class for platforms that have no enable mode.

    Netmiko translates the meaning of "enable" mode to be a proxy for "can
    go into config mode". In other words, that you ultimately have privileges
    to execute configuration changes.

    The expectation on platforms that have no method for elevating privileges
    is that the standard default privileges allow configuration changes.

    Consequently check_enable_mode returns True by default for platforms that
    don't explicitly support enable mode.
    """

    def check_enable_mode(self, check_string: str = "") -> bool:
        return True

    def enable(
        self,
        cmd: str = "",
        pattern: str = "",
        enable_pattern: Optional[str] = None,
        re_flags: int = re.IGNORECASE,
    ) -> str:
        return ""

    def exit_enable_mode(self, exit_command: str = "") -> str:
        return ""



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2014 - 2020 Kirk Byers
# Copyright (c) 2014 - 2020 Twin Bridges Technology
# Copyright (c) 2019 - 2020 NOKIA Inc.
# MIT License - See License file at:
#   https://github.com/ktbyers/netmiko/blob/develop/LICENSE

import re
import os
import time
from typing import Any, Optional, Union, Sequence, TextIO, Callable

from netmiko import log
from netmiko.base_connection import BaseConnection
from netmiko.scp_handler import BaseFileTransfer


class NokiaSros(BaseConnection):
    """
    Implement methods for interacting with Nokia SR OS devices
    for both SSH and telnet.

    Not applicable in Nokia SR OS (disabled):
        - exit_enable_mode()

    Overriden methods to adapt Nokia SR OS behavior (changed):
        - session_preparation()
        - set_base_prompt()
        - config_mode()
        - exit_config_mode()
        - check_config_mode()
        - save_config()
        - commit()
        - strip_prompt()
        - enable()
        - check_enable_mode()
    """

    def session_preparation(self) -> None:
        self._test_channel_read()
        self.set_base_prompt()
        # "@" indicates model-driven CLI (vs Classical CLI)
        if "@" in self.base_prompt:
            self._disable_complete_on_space()
            self.set_terminal_width(
                command="environment console width 512", pattern="environment"
            )
            self.disable_paging(command="environment more false")
            # To perform file operations we need to disable paging in classical-CLI also
            self.disable_paging(command="//environment no more")
        else:
            # Classical CLI has no method to set the terminal width nor to disable command
            # complete on space; consequently, cmd_verify needs disabled.
            self.global_cmd_verify = False
            self.disable_paging(command="environment no more", pattern="environment")

        # Clear the read buffer
        time.sleep(0.3 * self.global_delay_factor)
        self.clear_buffer()

    def set_base_prompt(self, *args: Any, **kwargs: Any) -> str:
        """Remove the > when navigating into the different config level."""
        cur_base_prompt = super().set_base_prompt(*args, **kwargs)
        match = re.search(r"\*?(.*?)(>.*)*#", cur_base_prompt)
        if match:
            # strip off >... from base_prompt; strip off leading *
            self.base_prompt: str = match.group(1)

        return self.base_prompt

    def _disable_complete_on_space(self) -> str:
        """
        SR-OS tries to auto complete commands when you type a "space" character.

        This is a bad idea for automation as what your program is sending no longer matches
        the command echo from the device, so we disable this behavior.
        """
        delay_factor = self.select_delay_factor(delay_factor=0)
        time.sleep(delay_factor * 0.1)
        command = "environment command-completion space false"
        self.write_channel(self.normalize_cmd(command))
        time.sleep(delay_factor * 0.1)
        return self.read_channel()

    def enable(
        self,
        cmd: str = "enable",
        pattern: str = "ssword",
        enable_pattern: Optional[str] = None,
        re_flags: int = re.IGNORECASE,
    ) -> str:
        """Enable SR OS administrative mode"""
        if "@" not in self.base_prompt:
            cmd = "enable-admin"
        return super().enable(cmd=cmd, pattern=pattern, re_flags=re_flags)

    def check_enable_mode(self, check_string: str = "in admin mode") -> bool:
        """Check if in enable mode."""
        cmd = "enable"
        if "@" not in self.base_prompt:
            cmd = "enable-admin"
        self.write_channel(self.normalize_cmd(cmd))
        output = self.read_until_prompt_or_pattern(
            pattern="ssword", read_entire_line=True
        )
        if "ssword" in output:
            self.write_channel(self.RETURN)  # send ENTER to pass the password prompt
            self.read_until_prompt(read_entire_line=True)
        return check_string in output

    def exit_enable_mode(self, *args: Any, **kwargs: Any) -> str:
        """Nokia SR OS does not have a notion of exiting administrative mode"""
        return ""

    def config_mode(
        self,
        config_command: str = "edit-config exclusive",
        pattern: str = "",
        re_flags: int = 0,
    ) -> str:
        """Enable config edit-mode for Nokia SR OS"""
        output = ""
        if not pattern:
            pattern = rf"\(ex\)\[.*{self.base_prompt}.*$"
            re_flags = re.DOTALL
        # Only model-driven CLI supports config-mode
        if "@" in self.base_prompt:
            output += super().config_mode(
                config_command=config_command, pattern=pattern, re_flags=re_flags
            )
        return output

    def exit_config_mode(self, *args: Any, **kwargs: Any) -> str:
        """Disable config edit-mode for Nokia SR OS"""
        output = self._exit_all()
        # Model-driven CLI
        if "@" in self.base_prompt and "(ex)[" in output:
            # Asterisk indicates changes were made.
            if "*(ex)[" in output:
                log.warning("Uncommitted changes! Discarding changes!")
                output += self._discard()
            cmd = "quit-config"
            self.write_channel(self.normalize_cmd(cmd))
            if self.global_cmd_verify is not False:
                output += self.read_until_pattern(pattern=re.escape(cmd))
                output += self.read_until_prompt(read_entire_line=True)
            else:
                output += self.read_until_prompt(read_entire_line=True)
        if self.check_config_mode():
            raise ValueError("Failed to exit configuration mode")
        return output

    def check_config_mode(
        self, check_string: str = r"(ex)[", pattern: str = r"@"
    ) -> bool:
        """Check config mode for Nokia SR OS"""
        if "@" not in self.base_prompt:
            # Classical CLI
            return False
        else:
            # Model-driven CLI look for "exclusive"
            return super().check_config_mode(check_string=check_string, pattern=pattern)

    def save_config(self, *args: Any, **kwargs: Any) -> str:
        """Persist configuration to cflash for Nokia SR OS"""
        return self._send_command_str(command_string="/admin save", expect_string=r"#")

    def send_config_set(
        self,
        config_commands: Union[str, Sequence[str], TextIO, None] = None,
        exit_config_mode: bool = None,
        **kwargs: Any,
    ) -> str:
        """Model driven CLI requires you not exit from configuration mode."""
        if exit_config_mode is None:
            # Set to False if model-driven CLI
            exit_config_mode = False if "@" in self.base_prompt else True
        return super().send_config_set(
            config_commands=config_commands, exit_config_mode=exit_config_mode, **kwargs
        )

    def commit(self, *args: Any, **kwargs: Any) -> str:
        """Activate changes from private candidate for Nokia SR OS"""
        output = self._exit_all()
        if "@" in self.base_prompt and "*(ex)[" in output:
            log.info("Apply uncommitted changes!")
            cmd = "commit"
            self.write_channel(self.normalize_cmd(cmd))
            new_output = ""
            if self.global_cmd_verify is not False:
                new_output += self.read_until_pattern(pattern=re.escape(cmd))
            if "@" not in new_output:
                new_output += self.read_until_pattern(r"@")
            output += new_output
        return output

    def _exit_all(self) -> str:
        """Return to the 'root' context."""
        output = ""
        exit_cmd = "exit all"
        self.write_channel(self.normalize_cmd(exit_cmd))
        # Make sure you read until you detect the command echo (avoid getting out of sync)
        if self.global_cmd_verify is not False:
            output += self.read_until_pattern(pattern=re.escape(exit_cmd))
            output += self.read_until_prompt(read_entire_line=True)
        else:
            output += self.read_until_prompt(read_entire_line=True)
        return output

    def _discard(self) -> str:
        """Discard changes from private candidate for Nokia SR OS"""
        output = ""
        if "@" in self.base_prompt:
            cmd = "discard"
            self.write_channel(self.normalize_cmd(cmd))
            new_output = ""
            if self.global_cmd_verify is not False:
                new_output += self.read_until_pattern(pattern=re.escape(cmd))
            if "@" not in new_output:
                new_output += self.read_until_prompt(read_entire_line=True)
            output += new_output
        return output

    def strip_prompt(self, *args: Any, **kwargs: Any) -> str:
        """Strip prompt from the output."""
        output = super().strip_prompt(*args, **kwargs)
        if "@" in self.base_prompt:
            # Remove context prompt too
            strips = r"[\r\n]*\!?\*?(\((ex|gl|pr|ro)\))?\[\S*\][\r\n]*"
            return re.sub(strips, "", output)
        else:
            return output

    def cleanup(self, command: str = "logout") -> None:
        """Gracefully exit the SSH session."""
        try:
            # The pattern="" forces use of send_command_timing
            if self.check_config_mode(pattern=""):
                self.exit_config_mode()
        except Exception:
            pass
        # Always try to send final 'logout'.
        self._session_log_fin = True
        self.write_channel(command + self.RETURN)


class NokiaSrosSSH(NokiaSros):
    """Nokia SR OS SSH driver."""

    pass


class NokiaSrosTelnet(NokiaSros):
    """Nokia SR OS Telnet driver."""

    pass


class NokiaSrosFileTransfer(BaseFileTransfer):
    def __init__(
        self,
        ssh_conn: BaseConnection,
        source_file: str,
        dest_file: str,
        file_system: Optional[str] = None,
        direction: str = "put",
        socket_timeout: float = 10.0,
        progress: Optional[Callable[..., Any]] = None,
        progress4: Optional[Callable[..., Any]] = None,
        hash_supported: bool = False,
    ) -> None:
        super().__init__(
            ssh_conn=ssh_conn,
            source_file=source_file,
            dest_file=dest_file,
            file_system=file_system,
            direction=direction,
            socket_timeout=socket_timeout,
            progress=progress,
            progress4=progress4,
            hash_supported=hash_supported,
        )

    def _file_cmd_prefix(self) -> str:
        """
        Allow MD-CLI to execute file operations by using classical CLI.

        Returns "//" if the current prompt is MD-CLI (empty string otherwise).
        """
        return "//" if "@" in self.ssh_ctl_chan.base_prompt else ""

    def remote_space_available(
        self, search_pattern: str = r"(\d+)\s+\w+\s+free"
    ) -> int:
        """Return space available on remote device."""

        # Sample text for search_pattern.
        # "               3 Dir(s)               961531904 bytes free."
        remote_cmd = self._file_cmd_prefix() + "file dir {}".format(self.file_system)
        remote_output = self.ssh_ctl_chan._send_command_str(remote_cmd)
        match = re.search(search_pattern, remote_output)
        assert match is not None
        return int(match.group(1))

    def check_file_exists(self, remote_cmd: str = "") -> bool:
        """Check if destination file exists (returns boolean)."""

        if self.direction == "put":
            if not remote_cmd:
                remote_cmd = self._file_cmd_prefix() + "file dir {}/{}".format(
                    self.file_system, self.dest_file
                )
            dest_file_name = self.dest_file.replace("\\", "/").split("/")[-1]
            remote_out = self.ssh_ctl_chan.send_command(remote_cmd)
            if "File Not Found" in remote_out:
                return False
            elif dest_file_name in remote_out:
                return True
            else:
                raise ValueError("Unexpected output from check_file_exists")
        elif self.direction == "get":
            return os.path.exists(self.dest_file)
        else:
            raise ValueError("Unexpected value for self.direction")

    def remote_file_size(
        self, remote_cmd: str = "", remote_file: Optional[str] = None
    ) -> int:
        """Get the file size of the remote file."""

        if remote_file is None:
            if self.direction == "put":
                remote_file = self.dest_file
            elif self.direction == "get":
                remote_file = self.source_file
            else:
                raise ValueError("Unexpected value for self.direction")
        if not remote_cmd:
            remote_cmd = self._file_cmd_prefix() + "file dir {}/{}".format(
                self.file_system, remote_file
            )
        remote_out = self.ssh_ctl_chan._send_command_str(remote_cmd)

        if "File Not Found" in remote_out:
            raise IOError("Unable to find file on remote system")

        dest_file_name = remote_file.replace("\\", "/").split("/")[-1]
        # Parse dir output for filename. Output format is:
        # "10/16/2019  10:00p                6738 {dest_file_name}"

        pattern = r"\S+\s+\S+\s+(\d+)\s+{}".format(re.escape(dest_file_name))
        match = re.search(pattern, remote_out)

        if not match:
            raise ValueError("Filename entry not found in dir output")

        file_size = int(match.group(1))
        return file_size

    def verify_file(self) -> bool:
        """Verify the file has been transferred correctly based on filesize."""
        if self.direction == "put":
            return os.stat(self.source_file).st_size == self.remote_file_size(
                remote_file=self.dest_file
            )
        elif self.direction == "get":
            return (
                self.remote_file_size(remote_file=self.source_file)
                == os.stat(self.dest_file).st_size
            )
        else:
            raise ValueError("Unexpected value of self.direction")

    def file_md5(self, file_name: str, add_newline: bool = False) -> str:
        raise AttributeError("SR-OS does not support an MD5-hash operation.")

    @staticmethod
    def process_md5(md5_output: str, pattern: str = "") -> str:
        raise AttributeError("SR-OS does not support an MD5-hash operation.")

    def compare_md5(self) -> bool:
        raise AttributeError("SR-OS does not support an MD5-hash operation.")

    def remote_md5(self, base_cmd: str = "", remote_file: Optional[str] = None) -> str:
        raise AttributeError("SR-OS does not support an MD5-hash operation.")



from netmiko.no_enable import NoEnable
from netmiko.no_config import NoConfig
from netmiko.base_connection import BaseConnection


class JuniperScreenOsSSH(NoEnable, NoConfig, BaseConnection):
    """
    Implement methods for interacting with Juniper ScreenOS devices.
    """

    def session_preparation(self) -> None:
        """
        ScreenOS can be configured to require: Accept this agreement y/[n]
        """
        terminator = r"\->"
        pattern = rf"(Accept this|{terminator})"
        data = self._test_channel_read(pattern=pattern)
        if "Accept this" in data:
            self.write_channel("y")
            data += self._test_channel_read(pattern=terminator)

        self.set_base_prompt()
        self.disable_paging(command="set console page 0")

    def save_config(
        self,
        cmd: str = "save config",
        confirm: bool = False,
        confirm_response: str = "",
    ) -> str:
        """Save Config."""
        return self._send_command_str(command_string=cmd)



from datetime import datetime
import pytest
import time
import re
from netmiko import ReadTimeout

# Setting to True will slow things down (as will wait until "show tech-support" completes.
CLEANUP = True


def my_cleanup(conn, sleep=180):
    try:
        # Must be long enough for "show tech-support" to finish
        time.sleep(sleep)
        conn.disconnect()
    except Exception:
        # If it fails, just try to keep going
        pass


def show_long_running(
    conn, read_timeout, expect_string, delay_factor=None, max_loops=None
):
    start_time = datetime.now()
    my_exception = None
    try:
        conn.send_command(
            "show tech-support",
            read_timeout=read_timeout,
            expect_string=expect_string,
            delay_factor=delay_factor,
            max_loops=max_loops,
        )
    except Exception as e:
        my_exception = e
    finally:
        end_time = datetime.now()
        exec_time = end_time - start_time
        # clean-up
        if CLEANUP:
            my_cleanup(conn)
        return (my_exception, exec_time)


def show_long_running_notimeout(conn, read_timeout):
    start_time = datetime.now()
    output = conn.send_command("show tech-support", read_timeout=read_timeout)
    end_time = datetime.now()
    exec_time = end_time - start_time
    return (output, exec_time)


def test_read_longrunning_cmd(net_connect_newconn):

    read_timeout = 300
    output, exec_time = show_long_running_notimeout(
        net_connect_newconn, read_timeout=read_timeout
    )
    assert "show interface" in output
    assert exec_time.total_seconds() > 10


@pytest.mark.parametrize(
    "test_timeout,allowed_percentage",
    [(0.4, 5.0), (1, 2.0), (5, 1.0), (10, 0.5), (60, 0.2)],
)
def test_read_timeout(net_connect_newconn, test_timeout, allowed_percentage):

    # Explicitly send expect_string so timing is more accurate
    my_prompt = net_connect_newconn.find_prompt()
    pattern = re.escape(my_prompt)

    my_except, exec_time = show_long_running(
        net_connect_newconn, read_timeout=test_timeout, expect_string=pattern
    )

    # Returned exception should be read_timeout
    assert isinstance(my_except, ReadTimeout)

    # Calculate difference in execution time from timeout
    time_diff = abs(exec_time.total_seconds() - test_timeout)
    # Convert difference to a percentage of expected timeout
    time_margin_percent = time_diff / test_timeout * 100
    # Margin off should be less than the allowed_percentage
    assert time_margin_percent < allowed_percentage


@pytest.mark.parametrize(
    "test_timeout,allowed_percentage", [(1, 2.0), (5, 1.0), (10, 0.5), (60, 0.2)]
)
def test_read_timeout_override(net_connect_newconn, test_timeout, allowed_percentage):

    net_connect_newconn.read_timeout_override = 12
    ssh_conn = net_connect_newconn

    # Explicitly send expect_string so timing is more accurate
    my_prompt = net_connect_newconn.find_prompt()
    pattern = re.escape(my_prompt)

    my_except, exec_time = show_long_running(
        ssh_conn, read_timeout=test_timeout, expect_string=pattern
    )

    # Returned exception should be read_timeout
    assert isinstance(my_except, ReadTimeout)

    # For comparsions compare to the override time with a fixed allowed percentage
    test_timeout = ssh_conn.read_timeout_override
    allowed_percentage = 0.5

    # Calculate difference in execution time from read_timeout_override
    time_diff = abs(exec_time.total_seconds() - test_timeout)

    # Convert difference to a percentage of expected timeout
    time_margin_percent = time_diff / test_timeout * 100
    # Margin off should be less than the allowed_percentage
    assert time_margin_percent < allowed_percentage


def test_deprecation_delay_factor(net_connect_newconn):

    ssh_conn = net_connect_newconn
    cmd = "show ip int brief"
    with pytest.deprecated_call():
        ssh_conn.send_command(cmd, delay_factor=2)


def test_deprecation_max_loops(net_connect_newconn):

    ssh_conn = net_connect_newconn
    cmd = "show ip int brief"
    with pytest.deprecated_call():
        ssh_conn.send_command(cmd, max_loops=1000)


@pytest.mark.parametrize(
    "fast_cli,delay_factor,max_loops,compat_timeout",
    [
        (True, None, None, 10),
        (False, 1.0, 100, 20),
        (False, 2.0, 100, 40),
        (False, 5.0, 40, 40),
    ],
)
def test_netmiko3_compatibility_mode(
    net_connect_newconn, fast_cli, delay_factor, max_loops, compat_timeout
):
    """
    In this test read_timeout is ignored and the timeout used by
    send_command should be calculated from global_delay_factor/delay_factor/max_loops
    """

    nc = net_connect_newconn
    # Force it into Netmiko 3.x compatibility mode
    nc.delay_factor_compat = True

    allowed_percentage = 0.5
    if not fast_cli:
        nc.fast_cli = False
        nc.global_delay_factor = 1.0

    # Explicitly send expect_string so timing is more accurate
    my_prompt = nc.find_prompt()
    pattern = re.escape(my_prompt)
    my_except, exec_time = show_long_running(
        nc,
        read_timeout=20,
        expect_string=pattern,
        delay_factor=delay_factor,
        max_loops=max_loops,
    )
    print(exec_time)

    # Returned exception should be read_timeout
    assert isinstance(my_except, ReadTimeout)

    # Calculate difference in execution time from timeout
    time_diff = abs(exec_time.total_seconds() - compat_timeout)
    # Convert difference to a percentage of expected timeout
    time_margin_percent = time_diff / compat_timeout * 100
    # Margin off should be less than the allowed_percentage
    assert time_margin_percent < allowed_percentage



from netmiko.cisco_base_connection import CiscoSSHConnection


class ExtremeWingSSH(CiscoSSHConnection):
    """Extreme WiNG support."""

    def session_preparation(self) -> None:
        """Disable paging and set Max term width"""
        self._test_channel_read(pattern=r">|#")
        self.set_base_prompt()
        self.set_terminal_width(command="terminal width 512", pattern="terminal")
        self.disable_paging(command="no page")



"""Mellanox MLNX-OS Switch support."""
import re
from typing import Optional

from netmiko.cisco_base_connection import CiscoSSHConnection
from netmiko import log


class MellanoxMlnxosSSH(CiscoSSHConnection):
    """Mellanox MLNX-OS Switch support."""

    def enable(
        self,
        cmd: str = "enable",
        pattern: str = "#",
        enable_pattern: Optional[str] = None,
        re_flags: int = re.IGNORECASE,
    ) -> str:
        """Enter into enable mode."""
        output = ""
        if not self.check_enable_mode():
            self.write_channel(self.normalize_cmd(cmd))
            output += self.read_until_prompt_or_pattern(
                pattern=pattern, re_flags=re_flags, read_entire_line=True
            )
            if not self.check_enable_mode():
                raise ValueError("Failed to enter enable mode.")
        return output

    def config_mode(
        self,
        config_command: str = "config term",
        pattern: str = r"\#",
        re_flags: int = 0,
    ) -> str:
        return super().config_mode(
            config_command=config_command, pattern=pattern, re_flags=re_flags
        )

    def check_config_mode(
        self, check_string: str = "(config", pattern: str = r"#"
    ) -> bool:
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def disable_paging(
        self,
        command: str = "no cli session paging enable",
        delay_factor: Optional[float] = None,
        cmd_verify: bool = True,
        pattern: Optional[str] = None,
    ) -> str:
        return super().disable_paging(
            command=command,
            delay_factor=delay_factor,
            cmd_verify=cmd_verify,
            pattern=pattern,
        )

    def exit_config_mode(self, exit_config: str = "exit", pattern: str = "#") -> str:
        """Mellanox does not support a single command to completely exit configuration mode.

        Consequently, need to keep checking and sending "exit".
        """
        output = ""
        check_count = 12
        while check_count >= 0:
            if self.check_config_mode():
                self.write_channel(self.normalize_cmd(exit_config))
                output += self.read_until_pattern(pattern=pattern)
            else:
                break
            check_count -= 1

        # One last check for whether we successfully exited config mode
        if self.check_config_mode():
            raise ValueError("Failed to exit configuration mode")

        log.debug(f"exit_config_mode: {output}")
        return output

    def save_config(
        self,
        cmd: str = "configuration write",
        confirm: bool = False,
        confirm_response: str = "",
    ) -> str:
        """Save Config on Mellanox devices Enters and Leaves Config Mode"""
        output = self.enable()
        output += self.config_mode()
        output += self._send_command_str(cmd)
        output += self.exit_config_mode()
        return output



#!/usr/bin/env python
import time
import hashlib
import io
import logging
from netmiko import ConnectHandler


def calc_md5(file_name=None, contents=None):
    """Compute MD5 hash of file."""
    if contents is not None:
        pass
    elif file_name:
        with open(file_name, "rb") as f:
            contents = f.read()
    else:
        raise ValueError("Most specify either file_name or contents")

    return hashlib.md5(contents.strip()).hexdigest()


def read_session_log(session_file, append=False):
    """Leading white-space can vary. Strip off leading white-space."""
    with open(session_file, "rb") as f:
        if append is True:
            line = f.readline().decode()
            assert "Initial file contents" in line
        log_content = f.read().lstrip()
        return log_content


def session_action(my_connect, command):
    """Common actions in the netmiko session to generate the session log."""
    time.sleep(1)
    my_connect.clear_buffer()
    output = my_connect.send_command(command)
    my_connect.disconnect()
    return output


def session_log_md5(session_file, compare_file):
    """Compare the session_log MD5 to the compare_file MD5"""
    compare_log_md5 = calc_md5(file_name=compare_file)
    log_content = read_session_log(session_file)
    session_log_md5 = calc_md5(contents=log_content)
    assert session_log_md5 == compare_log_md5


def session_log_md5_append(session_file, compare_file):
    """Compare the session_log MD5 to the compare_file MD5"""
    compare_log_md5 = calc_md5(file_name=compare_file)
    log_content = read_session_log(session_file, append=True)
    session_log_md5 = calc_md5(contents=log_content)
    assert session_log_md5 == compare_log_md5


def test_session_log(net_connect, commands, expected_responses):
    """Verify session_log matches expected content."""
    command = commands["basic"]
    session_action(net_connect, command)

    compare_file = expected_responses["compare_log"]
    session_file = expected_responses["session_log"]

    session_log_md5(session_file, compare_file)


def test_session_log_write(net_connect_slog_wr, commands, expected_responses):
    """Verify session_log matches expected content, but when channel writes are also logged."""
    command = commands["basic"]
    nc = net_connect_slog_wr

    # Send a marker down the channel
    nc.send_command("show foooooooo")
    time.sleep(1)
    nc.clear_buffer()
    nc.send_command(command)
    nc.disconnect()

    compare_file = expected_responses["compare_log_wr"]
    session_file = expected_responses["session_log_wr"]

    with open(compare_file, "rb") as f:
        compare_contents = f.read()

    # Header information varies too much due to device behavior differences.
    # So just discard it.
    marker = b"% Invalid input detected at '^' marker."
    _, compare_contents = compare_contents.split(marker)
    compare_log_md5 = calc_md5(contents=compare_contents.strip())

    log_content = read_session_log(session_file)
    marker = b"% Invalid input detected at '^' marker."
    _, log_content = log_content.split(marker)

    session_log_md5 = calc_md5(contents=log_content.strip())
    assert session_log_md5 == compare_log_md5


def test_session_log_append(device_slog, commands, expected_responses):
    """Verify session_log matches expected content, but when channel writes are also logged."""
    session_file = expected_responses["session_log_append"]
    # Create a starting file
    with open(session_file, "wb") as f:
        f.write(b"Initial file contents\n\n")

    # The netmiko connection has not been established yet.
    device_slog["session_log"] = session_file

    conn = ConnectHandler(**device_slog)
    command = commands["basic"]
    session_action(conn, command)

    compare_file = expected_responses["compare_log_append"]
    session_log_md5_append(session_file, compare_file)


def test_session_log_secrets(device_slog):
    """Verify session_log does not contain password or secret."""
    conn = ConnectHandler(**device_slog)
    conn.session_log.write("\nTesting password and secret replacement\n")
    conn.session_log.write("This is my password {}\n".format(conn.password))
    conn.session_log.write("This is my secret {}\n".format(conn.secret))

    file_name = device_slog["session_log"]
    with open(file_name, "r") as f:
        session_log = f.read()
    if conn.password:
        assert conn.password not in session_log
    if conn.secret:
        assert conn.secret not in session_log


def test_logging_filter_secrets(net_connect_slog_wr):
    """Verify logging DEBUG output does not contain password or secret."""

    nc = net_connect_slog_wr

    # setup logger to output to file
    file_name = "SLOG/netmiko.log"
    netmikologger = logging.getLogger("netmiko")
    netmikologger.setLevel(logging.DEBUG)
    file_handler = logging.FileHandler(file_name)
    file_handler.setLevel(logging.DEBUG)
    netmikologger.addHandler(file_handler)

    # cleanup the log file
    with open(file_name, "w") as f:
        f.write("")

    # run sequence
    nc.enable()
    time.sleep(1)
    nc.clear_buffer()
    nc.disconnect()

    with open(file_name, "r") as f:
        netmiko_log = f.read()
    if nc.password:
        assert nc.password not in netmiko_log
    if nc.secret:
        assert nc.secret not in netmiko_log


def test_unicode(device_slog):
    """Verify that you can write unicode characters into the session_log."""
    conn = ConnectHandler(**device_slog)

    smiley_face = "\N{grinning face with smiling eyes}"
    conn.session_log.write("\nTesting unicode\n")
    conn.session_log.write(smiley_face)
    conn.session_log.write(smiley_face)

    file_name = device_slog["session_log"]
    with open(file_name, "r") as f:
        session_log = f.read()
        assert smiley_face in session_log


def test_session_log_bytesio(device_slog, commands, expected_responses):
    """Verify session_log matches expected content, but when channel writes are also logged."""
    s_log = io.BytesIO()

    # The netmiko connection has not been established yet.
    device_slog["session_log"] = s_log
    device_slog["session_log_file_mode"] = "write"

    conn = ConnectHandler(**device_slog)
    command = commands["basic"]
    session_action(conn, command)

    compare_file = expected_responses["compare_log"]
    compare_log_md5 = calc_md5(file_name=compare_file)

    log_content = s_log.getvalue()
    session_log_md5 = calc_md5(contents=log_content)
    assert session_log_md5 == compare_log_md5



"""Dell N2/3/4000 base driver- supports DNOS6."""
from netmiko.dell.dell_powerconnect import DellPowerConnectBase


class DellDNOS6Base(DellPowerConnectBase):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self.ansi_escape_codes = True
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self.enable()
        self.set_terminal_width()
        self.disable_paging(command="terminal length 0")

    def save_config(
        self,
        cmd: str = "copy running-configuration startup-configuration",
        confirm: bool = False,
        confirm_response: str = "",
    ) -> str:
        """Saves Config"""
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )


class DellDNOS6SSH(DellDNOS6Base):
    pass


class DellDNOS6Telnet(DellDNOS6Base):
    pass



from typing import Any
from netmiko.no_enable import NoEnable
from netmiko.no_config import NoConfig
from netmiko.cisco_base_connection import CiscoSSHConnection


class BrocadeFOSSSH(NoEnable, NoConfig, CiscoSSHConnection):
    """Brocade Fabric OS support"""

    def __init__(self, **kwargs: Any) -> None:
        if kwargs.get("default_enter") is None:
            kwargs["default_enter"] = "\r"
        return super().__init__(**kwargs)

    def session_preparation(self) -> None:
        self._test_channel_read(pattern=r">")
        self.set_base_prompt()



from netmiko.ubiquiti.edge_ssh import UbiquitiEdgeSSH
from netmiko.ubiquiti.edgerouter_ssh import UbiquitiEdgeRouterSSH
from netmiko.ubiquiti.unifiswitch_ssh import UbiquitiUnifiSwitchSSH

__all__ = [
    "UbiquitiEdgeRouterSSH",
    "UbiquitiEdgeSSH",
    "UbiquitiUnifiSwitchSSH",
]



import time
import re
from typing import Any, Optional

from netmiko.base_connection import BaseConnection


class YamahaBase(BaseConnection):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self.disable_paging(command="console lines infinity")
        time.sleep(0.3 * self.global_delay_factor)
        self.clear_buffer()

    def check_enable_mode(self, check_string: str = "#") -> bool:
        return super().check_enable_mode(check_string=check_string)

    def enable(
        self,
        cmd: str = "administrator",
        pattern: str = r"Password",
        enable_pattern: Optional[str] = None,
        re_flags: int = re.IGNORECASE,
    ) -> str:
        return super().enable(
            cmd=cmd, pattern=pattern, enable_pattern=enable_pattern, re_flags=re_flags
        )

    def exit_enable_mode(self, exit_command: str = "exit") -> str:
        """
        When any changes have been made, the prompt 'Save new configuration ? (Y/N)'
        appears before exiting. Ignore this by entering 'N'.
        """
        output = ""
        if self.check_enable_mode():
            self.write_channel(self.normalize_cmd(exit_command))
            time.sleep(1)
            output = self.read_channel()
            if "(Y/N)" in output:
                self.write_channel("N")
            if self.base_prompt not in output:
                output += self.read_until_prompt(read_entire_line=True)
            if self.check_enable_mode():
                raise ValueError("Failed to exit enable mode.")
        return output

    def check_config_mode(self, check_string: str = "#", pattern: str = "") -> bool:
        """Checks if the device is in administrator mode or not."""
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def config_mode(
        self,
        config_command: str = "administrator",
        pattern: str = "Password",
        re_flags: int = re.IGNORECASE,
    ) -> str:
        return self.enable(cmd=config_command, pattern=pattern, re_flags=re_flags)

    def exit_config_mode(self, exit_config: str = "exit", pattern: str = ">") -> str:
        """
        No action taken. Call 'exit_enable_mode()' to explicitly exit Administration
        Level.
        """
        return ""

    def save_config(
        self, cmd: str = "save", confirm: bool = False, confirm_response: str = ""
    ) -> str:
        """Saves Config."""
        if confirm is True:
            raise ValueError("Yamaha does not support save_config confirmation.")
        self.enable()
        # Some devices are slow so match on trailing-prompt if you can
        return self._send_command_str(command_string=cmd)


class YamahaSSH(YamahaBase):
    """Yamaha SSH driver."""

    pass


class YamahaTelnet(YamahaBase):
    """Yamaha Telnet driver."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        default_enter = kwargs.get("default_enter")
        kwargs["default_enter"] = "\n" if default_enter is None else default_enter
        super().__init__(*args, **kwargs)



from typing import Any, Optional, Callable
import re
import os
from netmiko.base_connection import BaseConnection
from netmiko.cisco_base_connection import CiscoSSHConnection
from netmiko.cisco_base_connection import CiscoFileTransfer


class CiscoNxosSSH(CiscoSSHConnection):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self.ansi_escape_codes = True
        # NX-OS has an issue where it echoes the command even though it hasn't returned the prompt
        self._test_channel_read(pattern=r"[>#]")
        self.set_terminal_width(
            command="terminal width 511", pattern=r"terminal width 511"
        )
        self.disable_paging()
        self.set_base_prompt()

    def normalize_linefeeds(self, a_string: str) -> str:
        """Convert '\r\n' or '\r\r\n' to '\n, and remove extra '\r's in the text."""
        newline = re.compile(r"(\r\r\n\r|\r\r\n|\r\n)")
        # NX-OS fix for incorrect MD5 on 9K (due to strange <enter> patterns on NX-OS)
        return newline.sub(self.RESPONSE_RETURN, a_string).replace("\r", "\n")

    def check_config_mode(self, check_string: str = ")#", pattern: str = "#") -> bool:
        """Checks if the device is in configuration mode or not."""
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def save_config(
        self,
        cmd: str = "copy running-config startup-config",
        confirm: bool = False,
        confirm_response: str = "",
    ) -> str:
        self.enable()

        output = ""
        if confirm:
            output += self._send_command_timing_str(
                command_string=cmd, strip_prompt=False, strip_command=False
            )
            if confirm_response:
                output += self._send_command_timing_str(
                    confirm_response, strip_prompt=False, strip_command=False
                )
            else:
                # Send enter by default
                output += self._send_command_timing_str(
                    self.RETURN, strip_prompt=False, strip_command=False
                )
        else:
            # NX-OS is very slow on save_config ensure it waits long enough.
            output += self._send_command_str(
                command_string=cmd,
                strip_prompt=False,
                strip_command=False,
                read_timeout=100,
            )
        return output


class CiscoNxosFileTransfer(CiscoFileTransfer):
    """Cisco NXOS SCP File Transfer driver."""

    def __init__(
        self,
        ssh_conn: BaseConnection,
        source_file: str,
        dest_file: str,
        file_system: str = "bootflash:",
        direction: str = "put",
        socket_timeout: float = 10.0,
        progress: Optional[Callable[..., Any]] = None,
        progress4: Optional[Callable[..., Any]] = None,
        hash_supported: bool = True,
    ) -> None:
        self.ssh_ctl_chan = ssh_conn
        self.source_file = source_file
        self.dest_file = dest_file
        self.direction = direction

        if hash_supported is False:
            raise ValueError("hash_supported=False is not supported for NX-OS")

        if file_system:
            self.file_system = file_system
        else:
            raise ValueError("Destination file system must be specified for NX-OS")

        if direction == "put":
            self.source_md5 = self.file_md5(source_file)
            self.file_size = os.stat(source_file).st_size
        elif direction == "get":
            self.source_md5 = self.remote_md5(remote_file=source_file)
            self.file_size = self.remote_file_size(remote_file=source_file)
        else:
            raise ValueError("Invalid direction specified")

        self.socket_timeout = socket_timeout
        self.progress = progress
        self.progress4 = progress4

    def check_file_exists(self, remote_cmd: str = "") -> bool:
        """Check if the dest_file already exists on the file system (return boolean)."""
        if self.direction == "put":
            if not remote_cmd:
                remote_cmd = f"dir {self.file_system}{self.dest_file}"
            remote_out = self.ssh_ctl_chan._send_command_str(remote_cmd)
            search_string = r"{}.*Usage for".format(self.dest_file)
            if "No such file or directory" in remote_out:
                return False
            elif re.search(search_string, remote_out, flags=re.DOTALL):
                return True
            else:
                raise ValueError("Unexpected output from check_file_exists")
        elif self.direction == "get":
            return os.path.exists(self.dest_file)
        else:
            raise ValueError("Invalid value for file transfer direction.")

    def remote_file_size(
        self, remote_cmd: str = "", remote_file: Optional[str] = None
    ) -> int:
        """Get the file size of the remote file."""
        if remote_file is None:
            if self.direction == "put":
                remote_file = self.dest_file
            elif self.direction == "get":
                remote_file = self.source_file
            else:
                raise ValueError("Invalid value for file transfer direction.")

        if not remote_cmd:
            remote_cmd = f"dir {self.file_system}/{remote_file}"

        remote_out = self.ssh_ctl_chan._send_command_str(remote_cmd)
        if re.search("no such file or directory", remote_out, flags=re.I):
            raise IOError("Unable to find file on remote system")
        # Match line containing file name
        escape_file_name = re.escape(remote_file)
        pattern = r".*({}).*".format(escape_file_name)
        match = re.search(pattern, remote_out)
        if match:
            file_size = match.group(0)
            file_size = file_size.split()[0]
            return int(file_size)

        raise IOError("Unable to find file on remote system")

    @staticmethod
    def process_md5(md5_output: str, pattern: str = r"= (.*)") -> str:
        """Not needed on NX-OS."""
        raise NotImplementedError

    def remote_md5(
        self, base_cmd: str = "show file", remote_file: Optional[str] = None
    ) -> str:
        if remote_file is None:
            if self.direction == "put":
                remote_file = self.dest_file
            elif self.direction == "get":
                remote_file = self.source_file
        remote_md5_cmd = f"{base_cmd} {self.file_system}{remote_file} md5sum"
        output = self.ssh_ctl_chan._send_command_str(remote_md5_cmd, read_timeout=300)
        output = output.strip()
        return output

    def enable_scp(self, cmd: str = "") -> None:
        raise NotImplementedError

    def disable_scp(self, cmd: str = "") -> None:
        raise NotImplementedError



#!/usr/bin/env python
# CDOT = Centre for Development of Telematics, India
# CROS = CDOT Router OS
# Script: cros_ssh.py
# Author: Maloy Ghosh <mghosh@cdot.in>
# Updated by Kirk Byers
#
# Purpose: Provide basic SSH connection to CROS based router products

from typing import Optional, Union, Sequence, TextIO, Any
import time
import warnings
from netmiko.no_enable import NoEnable
from netmiko.cisco_base_connection import CiscoBaseConnection
from netmiko.base_connection import DELAY_FACTOR_DEPR_SIMPLE_MSG


class CdotCrosSSH(NoEnable, CiscoBaseConnection):
    """Implement methods for interacting with CROS network devices."""

    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read(pattern=r"[#\$]")
        self.set_base_prompt()
        self._disable_complete_on_space()
        self.set_terminal_width(command="screen-width 511", pattern=r"screen.width 511")
        self.disable_paging(command="screen-length 0")
        return

    def send_config_set(
        self,
        config_commands: Union[str, Sequence[str], TextIO, None] = None,
        exit_config_mode: bool = False,
        **kwargs: Any,
    ) -> str:
        """CROS requires you not exit from configuration mode."""
        return super().send_config_set(
            config_commands=config_commands, exit_config_mode=exit_config_mode, **kwargs
        )

    def check_config_mode(
        self, check_string: str = ")#", pattern: str = r"[#\$]"
    ) -> bool:
        """Checks if device is in configuration mode"""
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def config_mode(
        self, config_command: str = "config", pattern: str = "", re_flags: int = 0
    ) -> str:
        """Enter configuration mode."""
        return super().config_mode(
            config_command=config_command, pattern=pattern, re_flags=re_flags
        )

    def commit(
        self,
        comment: str = "",
        read_timeout: float = 120.0,
        delay_factor: Optional[float] = None,
        and_quit: bool = True,
    ) -> str:
        """
        Commit the candidate configuration.

        Commit the entered configuration. Raise an error and return the failure
        if the commit fails.

        default:
           command_string = commit
        comment:
           command_string = commit comment <comment>

        delay_factor: Deprecated in Netmiko 4.x. Will be eliminated in Netmiko 5.

        """

        if delay_factor is not None:
            warnings.warn(DELAY_FACTOR_DEPR_SIMPLE_MSG, DeprecationWarning)

        command_string = "commit"
        commit_marker = ["Commit complete", "No modifications to commit"]

        if comment:
            if '"' in comment:
                raise ValueError("Invalid comment contains double quote")
            command_string += f' comment "{comment}"'

        output = self.config_mode()
        output += self._send_command_str(
            command_string,
            strip_prompt=False,
            strip_command=True,
            read_timeout=read_timeout,
        )

        if not (any(x in output for x in commit_marker)):
            raise ValueError(f"Commit failed with the following errors:\n\n{output}")
        if and_quit:
            self.exit_config_mode()
        return output

    def _disable_complete_on_space(self) -> str:
        """
        CROS tries to auto complete commands when you type a "space" character.

        This is a bad idea for automation as what your program is sending no longer matches
        the command echo from the device. So we disable this behavior.
        """
        delay_factor = self.select_delay_factor(delay_factor=0)
        time.sleep(delay_factor * 0.1)
        command = "complete-on-space false"
        self.write_channel(self.normalize_cmd(command))
        time.sleep(delay_factor * 0.1)
        output = self.read_channel()
        return output



from typing import Any
from netmiko.cisco_base_connection import CiscoSSHConnection


class ExtremeNetironBase(CiscoSSHConnection):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self.disable_paging(command="skip-page-display")
        self.set_terminal_width()

    def save_config(
        self,
        cmd: str = "write memory",
        confirm: bool = False,
        confirm_response: str = "",
    ) -> str:
        """Save Config"""
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )


class ExtremeNetironSSH(ExtremeNetironBase):
    pass


class ExtremeNetironTelnet(ExtremeNetironBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        default_enter = kwargs.get("default_enter")
        kwargs["default_enter"] = "\r\n" if default_enter is None else default_enter
        super().__init__(*args, **kwargs)



#!/usr/bin/env python
import pytest
from netmiko import ConfigInvalidException
from netmiko import ReadTimeout


def test_ssh_connect(net_connect, commands, expected_responses):
    """
    Verify the connection was established successfully
    """
    show_version = net_connect.send_command(commands["version"])
    assert expected_responses["version_banner"] in show_version


def test_enable_mode(net_connect, commands, expected_responses):
    """
    Test entering enable mode

    Catch exception for devices that don't support enable
    """
    try:
        net_connect.enable()
        enable_prompt = net_connect.find_prompt()
        assert enable_prompt == expected_responses["enable_prompt"]
    except AttributeError:
        assert True


def test_config_mode(net_connect, commands, expected_responses):
    """
    Test enter config mode
    """
    # Behavior for devices with no config mode is to return null string
    if net_connect.config_mode() != "":
        assert net_connect.check_config_mode() is True
    else:
        pytest.skip("Platform doesn't support config mode.")


def test_exit_config_mode(net_connect, commands, expected_responses):
    """Test exit config mode."""
    if net_connect._config_mode:
        net_connect.exit_config_mode()
        assert net_connect.check_config_mode() is False
    else:
        pytest.skip("Platform doesn't support config mode.")


def test_config_set(net_connect, commands, expected_responses):
    """Test sending configuration commands."""

    config_commands = commands["config"]
    support_commit = commands.get("support_commit")
    config_verify = commands["config_verification"]

    # Set to initial value and testing sending command as a string
    net_connect.send_config_set(config_commands[0])
    if support_commit:
        net_connect.commit()
    cmd_response = expected_responses.get("cmd_response_init")
    config_commands_output = net_connect.send_command(config_verify)
    if cmd_response:
        assert cmd_response in config_commands_output
    else:
        assert config_commands[0] in config_commands_output
    net_connect.send_config_set(config_commands)
    if support_commit:
        net_connect.commit()
    cmd_response = expected_responses.get("cmd_response_final")
    config_commands_output = net_connect.send_command_expect(config_verify)
    if cmd_response:
        assert cmd_response in config_commands_output
    else:
        assert config_commands[-1] in config_commands_output


def test_config_set_longcommand(net_connect, commands, expected_responses):
    """Test sending configuration commands using long commands"""
    config_commands = commands.get("config_long_command")
    config_verify = commands["config_verification"]  # noqa
    if not config_commands:
        assert True
        return
    output = net_connect.send_config_set(config_commands)  # noqa
    assert True


def test_config_hostname(net_connect, commands, expected_responses):
    hostname = "test-netmiko1"
    command = f"hostname {hostname}"
    if "arista" in net_connect.device_type:
        current_hostname = net_connect.find_prompt()[:-1]
        net_connect.send_config_set(command)
        new_hostname = net_connect.find_prompt()
        assert hostname in new_hostname

        # Reset prompt back to original value
        net_connect.set_base_prompt()
        net_connect.send_config_set(f"hostname {current_hostname}")
        net_connect.set_base_prompt()


def test_config_from_file(net_connect, commands, expected_responses):
    """
    Test sending configuration commands from a file
    """
    config_file = commands.get("config_file")
    config_verify = commands["config_verification"]
    if config_file is not None:
        net_connect.send_config_from_file(config_file)
        config_commands_output = net_connect.send_command_expect(config_verify)
        assert expected_responses["file_check_cmd"] in config_commands_output
    else:
        assert pytest.skip()

    if "nokia_sros" in net_connect.device_type:
        net_connect.save_config()


def test_config_error_pattern(net_connect, commands, expected_responses):
    """
    Raise exception when config_error_str is present in output
    """
    error_pattern = commands.get("error_pattern")
    if error_pattern is None:
        pytest.skip("No error_pattern defined.")
    config_base = commands.get("config")
    config_err = commands.get("invalid_config")
    config_list = config_base + [config_err]

    # Should not raise an exception since error_pattern not specified
    net_connect.send_config_set(config_commands=config_list)

    if config_list and error_pattern:
        with pytest.raises(ConfigInvalidException):
            net_connect.send_config_set(
                config_commands=config_list, error_pattern=error_pattern
            )

        # Try it with cmd_verify=True also
        with pytest.raises(ConfigInvalidException):
            net_connect.send_config_set(
                config_commands=config_list,
                error_pattern=error_pattern,
                cmd_verify=True,
            )

    else:
        print("Skipping test: no error_pattern supplied.")


def test_banner(net_connect, commands, expected_responses):
    """
    Banner configuration has a special exclusing where cmd_verify is dynamically
    disabled so make sure it works.
    """
    # Make sure banner comes in as separate lines
    banner = commands.get("banner")
    if banner is None:
        pytest.skip("No banner defined.")
    # Make sure banner comes in as separate lines
    banner = banner.splitlines()
    config_base = commands.get("config")
    config_list = config_base + banner

    # Remove any existing banner
    net_connect.send_config_set("no banner login")

    # bypass_commands="" should fail as cmd_verify will be True
    with pytest.raises(ReadTimeout) as e:  # noqa
        net_connect.send_config_set(config_commands=config_list, bypass_commands="")

    # Recover from send_config_set failure. The "%" is to finish the failed banner.
    net_connect.write_channel("%\n")
    net_connect.exit_config_mode()

    net_connect.send_config_set(config_commands=config_list)
    show_run = net_connect.send_command("show run | inc banner log")
    assert "banner login" in show_run

    net_connect.send_config_set("no banner login")


def test_global_cmd_verify(net_connect, commands, expected_responses):
    """
    Banner configuration has a special exclusing where cmd_verify is dynamically
    disabled so make sure it works.
    """

    # Make sure banner comes in as separate lines
    banner = commands.get("banner")
    if banner is None:
        pytest.skip("No banner defined.")
    # Make sure banner comes in as separate lines
    banner = banner.splitlines()
    config_base = commands.get("config")
    config_list = config_base + banner

    # Remove any existing banner
    net_connect.send_config_set("no banner login")

    # bypass_commands="" should fail as cmd_verify will be True
    with pytest.raises(ReadTimeout) as e:  # noqa
        net_connect.send_config_set(config_commands=config_list, bypass_commands="")

    # Recover from send_config_set failure. The "%" is to finish the failed banner.
    net_connect.write_channel("%\n")
    net_connect.exit_config_mode()

    net_connect.global_cmd_verify = False
    # Should work now as global_cmd_verify is False
    net_connect.send_config_set(config_commands=config_list, bypass_commands="")
    show_run = net_connect.send_command("show run | inc banner log")
    assert "banner login" in show_run

    net_connect.send_config_set("no banner login")


def test_disconnect(net_connect, commands, expected_responses):
    """
    Terminate the SSH session
    """
    net_connect.disconnect()



"""
The ssh_autodetect module is used to auto-detect the netmiko device_type to use to further initiate
a new SSH connection with a remote host. This auto-detection is based on a unique class called
**SSHDetect**.

Notes
-----

The **SSHDetect** class is instantiated using the same parameters than a standard Netmiko
connection (see the *netmiko.ssh_dispatacher.ConnectHandler* function). The only acceptable value
for the 'device_type' argument is 'autodetect'.

The auto-detection is solely based on *SSH_MAPPER_BASE*. The keys are the name of
the 'device_type' supported for auto-detection and the value is another dictionary describing how
to handle the auto-detection.

* "cmd" : The command to send to the remote device. **The command output must not require paging.**
* "search_patterns" : A list of regex to compare with the output of the command
* "priority" : An integer (0-99) which specifies the confidence of the match above
* "dispatch" : The function to call to try the autodetection (per default SSHDetect._autodetect_std)

Examples
--------

# Auto-detection section
>>> from netmiko.ssh_autodetect import SSHDetect
>>> from netmiko.ssh_dispatcher import ConnectHandler
>>> remote_device = {'device_type': 'autodetect',
                     'host': 'remote.host',
                     'username': 'test',
                     'password': 'foo'}
>>> guesser = SSHDetect(**remote_device)
>>> best_match = guesser.autodetect()
>>> print(best_match) # Name of the best device_type to use further
>>> print(guesser.potential_matches) # Dictionary of the whole matching result

# Netmiko connection creation section
>>> remote_device['device_type'] = best_match
>>> connection = ConnectHandler(**remote_device)
"""
from typing import Any, List, Optional, Union, Dict
import re
import time

import paramiko

from netmiko.ssh_dispatcher import ConnectHandler
from netmiko.base_connection import BaseConnection


# 'dispatch' key is the SSHDetect method to call. dispatch key will be popped off dictionary
# remaining keys indicate kwargs that will be passed to dispatch method.
# Note, the 'cmd' needs to avoid output paging.
SSH_MAPPER_DICT = {
    "alcatel_aos": {
        "cmd": "show system",
        "search_patterns": [r"Alcatel-Lucent"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "alcatel_sros": {
        "cmd": "show version",
        "search_patterns": ["Nokia", "Alcatel"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "apresia_aeos": {
        "cmd": "show system",
        "search_patterns": ["Apresia"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "arista_eos": {
        "cmd": "show version",
        "search_patterns": [r"Arista"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "ciena_saos": {
        "cmd": "software show",
        "search_patterns": [r"saos"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "cisco_asa": {
        "cmd": "show version",
        "search_patterns": [r"Cisco Adaptive Security Appliance", r"Cisco ASA"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "cisco_ios": {
        "cmd": "show version",
        "search_patterns": [
            "Cisco IOS Software",
            "Cisco Internetwork Operating System Software",
        ],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "cisco_nxos": {
        "cmd": "show version",
        "search_patterns": [r"Cisco Nexus Operating System", r"NX-OS"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "cisco_xr": {
        "cmd": "show version",
        "search_patterns": [r"Cisco IOS XR"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "dell_force10": {
        "cmd": "show version",
        "search_patterns": [r"Real Time Operating System Software"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "dell_os9": {
        "cmd": "show system",
        "search_patterns": [
            r"Dell Application Software Version:  9",
            r"Dell Networking OS Version : 9",
        ],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "dell_os10": {
        "cmd": "show version",
        "search_patterns": [r"Dell EMC Networking OS10.Enterprise"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "dell_powerconnect": {
        "cmd": "show system",
        "search_patterns": [r"PowerConnect"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "f5_tmsh": {
        "cmd": "show sys version",
        "search_patterns": [r"BIG-IP"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "f5_linux": {
        "cmd": "cat /etc/issue",
        "search_patterns": [r"BIG-IP"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "hp_comware": {
        "cmd": "display version",
        "search_patterns": ["HPE Comware", "HP Comware"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "huawei": {
        "cmd": "display version",
        "search_patterns": [
            r"Huawei Technologies",
            r"Huawei Versatile Routing Platform Software",
        ],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "juniper_junos": {
        "cmd": "show version",
        "search_patterns": [
            r"JUNOS Software Release",
            r"JUNOS .+ Software",
            r"JUNOS OS Kernel",
            r"JUNOS Base Version",
        ],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "linux": {
        "cmd": "uname -a",
        "search_patterns": [r"Linux"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "extreme_exos": {
        "cmd": "show version",
        "search_patterns": [r"ExtremeXOS"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "extreme_netiron": {
        "cmd": "show version",
        "search_patterns": [r"(NetIron|MLX)"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "extreme_slx": {
        "cmd": "show version",
        "search_patterns": [r"SLX-OS Operating System Software"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "extreme_tierra": {
        "cmd": "show version",
        "search_patterns": [r"TierraOS Software"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "ubiquiti_edgeswitch": {
        "cmd": "show version",
        "search_patterns": [r"EdgeSwitch"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "cisco_wlc": {
        "cmd": "",
        "dispatch": "_autodetect_remote_version",
        "search_patterns": [r"CISCO_WLC"],
        "priority": 99,
    },
    "cisco_wlc_85": {
        "cmd": "show inventory",
        "dispatch": "_autodetect_std",
        "search_patterns": [r"Cisco Wireless Controller"],
        "priority": 99,
    },
    "mellanox_mlnxos": {
        "cmd": "show version",
        "search_patterns": [r"Onyx", r"SX_PPC_M460EX"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "yamaha": {
        "cmd": "show copyright",
        "search_patterns": [r"Yamaha Corporation"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "fortinet": {
        "cmd": "get system status",
        "search_patterns": [r"FortiOS"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "paloalto_panos": {
        "cmd": "show system info",
        "search_patterns": [r"model:\s+PA"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
    "supermicro_smis": {
        "cmd": "show system info",
        "search_patterns": [r"Super Micro Computer"],
        "priority": 99,
        "dispatch": "_autodetect_std",
    },
}

# Sort SSH_MAPPER_DICT such that the most common commands are first
cmd_count: Dict[str, int] = {}
for k, v in SSH_MAPPER_DICT.items():
    my_cmd = v["cmd"]
    assert isinstance(my_cmd, str)
    count = cmd_count.setdefault(my_cmd, 0)
    cmd_count[my_cmd] = count + 1
cmd_count = {k: v for k, v in sorted(cmd_count.items(), key=lambda item: item[1])}

# SSH_MAPPER_BASE is a list
SSH_MAPPER_BASE = sorted(
    SSH_MAPPER_DICT.items(), key=lambda item: int(cmd_count[str(item[1]["cmd"])])
)
SSH_MAPPER_BASE.reverse()


class SSHDetect(object):
    """
    The SSHDetect class tries to automatically guess the device type running on the SSH remote end.
    Be careful that the kwargs 'device_type' must be set to 'autodetect', otherwise it won't work at
    all.

    Parameters
    ----------
    *args : list
        The same *args that you might provide to the netmiko.ssh_dispatcher.ConnectHandler.
    *kwargs : dict
        The same *kwargs that you might provide to the netmiko.ssh_dispatcher.ConnectHandler.

    Attributes
    ----------
    connection : netmiko.terminal_server.TerminalServerSSH
        A basic connection to the remote SSH end.
    potential_matches: dict
        Dict of (device_type, accuracy) that is populated through an interaction with the
        remote end.

    Methods
    -------
    autodetect()
        Try to determine the device type.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Constructor of the SSHDetect class
        """
        if kwargs["device_type"] != "autodetect":
            raise ValueError("The connection device_type must be 'autodetect'")
        # Always set cmd_verify to False for autodetect
        kwargs["global_cmd_verify"] = False
        self.connection = ConnectHandler(*args, **kwargs)
        # Call the _test_channel_read() in base to clear initial data
        output = BaseConnection._test_channel_read(self.connection)
        self.initial_buffer = output
        self.potential_matches: Dict[str, int] = {}
        self._results_cache: Dict[str, str] = {}

    def autodetect(self) -> Union[str, None]:
        """
        Try to guess the best 'device_type' based on patterns defined in SSH_MAPPER_BASE

        Returns
        -------
        best_match : str or None
            The device type that is currently the best to use to interact with the device
        """
        for device_type, autodetect_dict in SSH_MAPPER_BASE:
            tmp_dict = autodetect_dict.copy()
            call_method = tmp_dict.pop("dispatch")
            assert isinstance(call_method, str)
            autodetect_method = getattr(self, call_method)
            accuracy = autodetect_method(**tmp_dict)
            if accuracy:
                self.potential_matches[device_type] = accuracy
                if accuracy >= 99:  # Stop the loop as we are sure of our match
                    best_match = sorted(
                        self.potential_matches.items(), key=lambda t: t[1], reverse=True
                    )
                    # WLC needs two different auto-dectect solutions
                    if "cisco_wlc_85" in best_match[0]:
                        best_match[0] = ("cisco_wlc", 99)

                    self.connection.disconnect()
                    return best_match[0][0]

        if not self.potential_matches:
            self.connection.disconnect()
            return None

        best_match = sorted(
            self.potential_matches.items(), key=lambda t: t[1], reverse=True
        )
        self.connection.disconnect()
        return best_match[0][0]

    def _send_command(self, cmd: str = "") -> str:
        """
        Handle reading/writing channel directly. It is also sanitizing the output received.

        Parameters
        ----------
        cmd : str, optional
            The command to send to the remote device (default : "", just send a new line)

        Returns
        -------
        output : str
            The output from the command sent
        """
        self.connection.write_channel(cmd + "\n")
        time.sleep(1)
        output = self.connection.read_channel_timing()
        output = self.connection.strip_backspaces(output)
        return output

    def _send_command_wrapper(self, cmd: str) -> str:
        """
        Send command to the remote device with a caching feature to avoid sending the same command
        twice based on the SSH_MAPPER_BASE dict cmd key.

        Parameters
        ----------
        cmd : str
            The command to send to the remote device after checking cache.

        Returns
        -------
        response : str
            The response from the remote device.
        """
        cached_results = self._results_cache.get(cmd)
        if not cached_results:
            response = self._send_command(cmd)
            self._results_cache[cmd] = response
            return response
        else:
            return cached_results

    def _autodetect_remote_version(
        self,
        search_patterns: Optional[List[str]] = None,
        re_flags: int = re.IGNORECASE,
        priority: int = 99,
        **kwargs: Any
    ) -> int:
        """
        Method to try auto-detect the device type, by matching a regular expression on the reported
        remote version of the SSH server.

        Parameters
        ----------
        search_patterns : list
            A list of regular expression to look for in the reported remote SSH version
            (default: None).
        re_flags: re.flags, optional
            Any flags from the python re module to modify the regular expression (default: re.I).
        priority: int, optional
            The confidence the match is right between 0 and 99 (default: 99).
        """
        invalid_responses = [r"^$"]

        if not search_patterns:
            return 0

        try:
            remote_conn = self.connection.remote_conn
            assert isinstance(remote_conn, paramiko.Channel)
            assert remote_conn.transport is not None
            remote_version = remote_conn.transport.remote_version
            for pattern in invalid_responses:
                match = re.search(pattern, remote_version, flags=re.I)
                if match:
                    return 0
            for pattern in search_patterns:
                match = re.search(pattern, remote_version, flags=re_flags)
                if match:
                    return priority
        except Exception:
            return 0
        return 0

    def _autodetect_std(
        self,
        cmd: str = "",
        search_patterns: Optional[List[str]] = None,
        re_flags: int = re.IGNORECASE,
        priority: int = 99,
    ) -> int:
        """
        Standard method to try to auto-detect the device type. This method will be called for each
        device_type present in SSH_MAPPER_BASE dict ('dispatch' key). It will attempt to send a
        command and match some regular expression from the ouput for each entry in SSH_MAPPER_BASE
        ('cmd' and 'search_pattern' keys).

        Parameters
        ----------
        cmd : str
            The command to send to the remote device after checking cache.
        search_patterns : list
            A list of regular expression to look for in the command's output (default: None).
        re_flags: re.flags, optional
            Any flags from the python re module to modify the regular expression (default: re.I).
        priority: int, optional
            The confidence the match is right between 0 and 99 (default: 99).
        """
        invalid_responses = [
            r"% Invalid input detected",
            r"syntax error, expecting",
            r"Error: Unrecognized command",
            r"%Error",
            r"command not found",
            r"Syntax Error: unexpected argument",
            r"% Unrecognized command found at",
        ]
        if not cmd or not search_patterns:
            return 0
        try:
            # _send_command_wrapper will use already cached results if available
            response = self._send_command_wrapper(cmd)
            # Look for error conditions in output
            for pattern in invalid_responses:
                match = re.search(pattern, response, flags=re.I)
                if match:
                    return 0
            for pattern in search_patterns:
                match = re.search(pattern, response, flags=re_flags)
                if match:
                    return priority
        except Exception:
            return 0
        return 0



import time
import re
from typing import Optional

from netmiko.cisco_base_connection import CiscoBaseConnection
from netmiko import log


class HuaweiSmartAXSSH(CiscoBaseConnection):
    """Supports Huawei SmartAX and OLT."""

    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self.ansi_escape_codes = True
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self._disable_smart_interaction()
        self.disable_paging()

    def strip_ansi_escape_codes(self, string_buffer: str) -> str:
        """
        Huawei does a strange thing where they add a space and then add ESC[1D
        to move the cursor to the left one.
        The extra space is problematic.
        """
        code_cursor_left = chr(27) + r"\[\d+D"
        output = string_buffer
        pattern = rf" {code_cursor_left}"
        output = re.sub(pattern, "", output)

        log.debug("Stripping ANSI escape codes")
        log.debug(f"new_output = {output}")
        log.debug(f"repr = {repr(output)}")
        return super().strip_ansi_escape_codes(output)

    def _disable_smart_interaction(
        self, command: str = "undo smart", delay_factor: float = 1.0
    ) -> None:
        """Disables the { <cr> } prompt to avoid having to sent a 2nd return after each command"""
        delay_factor = self.select_delay_factor(delay_factor)
        time.sleep(delay_factor * 0.1)
        self.clear_buffer()
        command = self.normalize_cmd(command)
        log.debug("In disable_smart_interaction")
        log.debug(f"Command: {command}")
        self.write_channel(command)
        if self.global_cmd_verify is not False:
            output = self.read_until_pattern(pattern=re.escape(command.strip()))
            output = self.read_until_prompt(read_entire_line=True)
        else:
            output = self.read_until_prompt(read_entire_line=True)
        log.debug(f"{output}")
        log.debug("Exiting disable_smart_interaction")

    def disable_paging(
        self,
        command: str = "scroll",
        delay_factor: Optional[float] = None,
        cmd_verify: bool = True,
        pattern: Optional[str] = None,
    ) -> str:
        return super().disable_paging(
            command=command,
            delay_factor=delay_factor,
            cmd_verify=cmd_verify,
            pattern=pattern,
        )

    def config_mode(
        self, config_command: str = "config", pattern: str = "", re_flags: int = 0
    ) -> str:
        return super().config_mode(
            config_command=config_command, pattern=pattern, re_flags=re_flags
        )

    def check_config_mode(self, check_string: str = ")#", pattern: str = "") -> bool:
        return super().check_config_mode(check_string=check_string)

    def exit_config_mode(
        self, exit_config: str = "return", pattern: str = r"#.*"
    ) -> str:
        return super().exit_config_mode(exit_config=exit_config)

    def check_enable_mode(self, check_string: str = "#") -> bool:
        return super().check_enable_mode(check_string=check_string)

    def enable(
        self,
        cmd: str = "enable",
        pattern: str = "",
        enable_pattern: Optional[str] = None,
        re_flags: int = re.IGNORECASE,
    ) -> str:
        return super().enable(
            cmd=cmd, pattern=pattern, re_flags=re_flags, enable_pattern=enable_pattern
        )

    def set_base_prompt(
        self,
        pri_prompt_terminator: str = ">",
        alt_prompt_terminator: str = "#",
        delay_factor: float = 1.0,
        pattern: Optional[str] = None,
    ) -> str:
        return super().set_base_prompt(
            pri_prompt_terminator=pri_prompt_terminator,
            alt_prompt_terminator=alt_prompt_terminator,
            delay_factor=delay_factor,
            pattern=pattern,
        )

    def save_config(
        self, cmd: str = "save", confirm: bool = False, confirm_response: str = ""
    ) -> str:
        """Save Config for HuaweiSSH"""
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )



"""
This module is used to auto-detect the type of a device in order to automatically create a
Netmiko connection.

The will avoid to hard coding the 'device_type' when using the ConnectHandler factory function
from Netmiko.

Example:
------------------
from netmiko.snmp_autodetect import SNMPDetect

my_snmp = SNMPDetect(hostname='1.1.1.70', user='pysnmp', auth_key='key1', encrypt_key='key2')
device_type = my_snmp.autodetect()
------------------

autodetect will return None if no match.

SNMPDetect class defaults to SNMPv3

Note, pysnmp is a required dependency for SNMPDetect and is intentionally not included in
netmiko requirements. So installation of pysnmp might be required.
"""
from typing import Optional, Dict
from typing.re import Pattern
import re

try:
    from pysnmp.entity.rfc3413.oneliner import cmdgen
except ImportError:
    raise ImportError("pysnmp not installed; please install it: 'pip install pysnmp'")

from netmiko.ssh_dispatcher import CLASS_MAPPER


# Higher priority indicates a better match.
SNMP_MAPPER_BASE = {
    "arista_eos": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*Arista Networks EOS.*", re.IGNORECASE),
        "priority": 99,
    },
    "paloalto_panos": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*Palo Alto Networks.*", re.IGNORECASE),
        "priority": 99,
    },
    "hp_comware": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*HP(E)? Comware.*", re.IGNORECASE),
        "priority": 99,
    },
    "hp_procurve": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".ProCurve", re.IGNORECASE),
        "priority": 99,
    },
    "cisco_ios": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*Cisco IOS Software,.*", re.IGNORECASE),
        "priority": 60,
    },
    "cisco_xe": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*IOS-XE Software,.*", re.IGNORECASE),
        "priority": 99,
    },
    "cisco_xr": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*Cisco IOS XR Software.*", re.IGNORECASE),
        "priority": 99,
    },
    "cisco_asa": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*Cisco Adaptive Security Appliance.*", re.IGNORECASE),
        "priority": 99,
    },
    "cisco_nxos": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*Cisco NX-OS.*", re.IGNORECASE),
        "priority": 99,
    },
    "cisco_wlc": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*Cisco Controller.*", re.IGNORECASE),
        "priority": 99,
    },
    "f5_tmsh": {
        "oid": ".1.3.6.1.4.1.3375.2.1.4.1.0",
        "expr": re.compile(r".*BIG-IP.*", re.IGNORECASE),
        "priority": 99,
    },
    "fortinet": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r"Forti.*", re.IGNORECASE),
        "priority": 80,
    },
    "checkpoint": {
        "oid": ".1.3.6.1.4.1.2620.1.6.16.9.0",
        "expr": re.compile(r"CheckPoint"),
        "priority": 79,
    },
    "juniper_junos": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*Juniper.*"),
        "priority": 99,
    },
    "nokia_sros": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r".*TiMOS.*"),
        "priority": 99,
    },
    "dell_powerconnect": {
        "oid": ".1.3.6.1.2.1.1.1.0",
        "expr": re.compile(r"PowerConnect.*", re.IGNORECASE),
        "priority": 50,
    },
}

# Ensure all SNMP device types are supported by Netmiko
SNMP_MAPPER = {}
std_device_types = list(CLASS_MAPPER.keys())
for device_type in std_device_types:
    if SNMP_MAPPER_BASE.get(device_type):
        SNMP_MAPPER[device_type] = SNMP_MAPPER_BASE[device_type]


class SNMPDetect(object):
    """
    The SNMPDetect class tries to automatically determine the device type.

    Typically this will use the MIB-2 SysDescr and regular expressions.

    Parameters
    ----------
    hostname: str
        The name or IP address of the hostname we want to guess the type
    snmp_version : str, optional ('v1', 'v2c' or 'v3')
        The SNMP version that is running on the device (default: 'v3')
    snmp_port : int, optional
        The UDP port on which SNMP is listening (default: 161)
    community : str, optional
        The SNMP read community when using SNMPv2 (default: None)
    user : str, optional
        The SNMPv3 user for authentication (default: '')
    auth_key : str, optional
        The SNMPv3 authentication key (default: '')
    encrypt_key : str, optional
        The SNMPv3 encryption key (default: '')
    auth_proto : str, optional ('des', '3des', 'aes128', 'aes192', 'aes256')
        The SNMPv3 authentication protocol (default: 'aes128')
    encrypt_proto : str, optional ('sha', 'md5')
        The SNMPv3 encryption protocol (default: 'sha')

    Attributes
    ----------
    hostname: str
        The name or IP address of the device we want to guess the type
    snmp_version : str
        The SNMP version that is running on the device
    snmp_port : int
        The UDP port on which SNMP is listening
    community : str
        The SNMP read community when using SNMPv2
    user : str
        The SNMPv3 user for authentication
    auth_key : str
        The SNMPv3 authentication key
    encrypt_key : str
        The SNMPv3 encryption key
    auth_proto : str
        The SNMPv3 authentication protocol
    encrypt_proto : str
        The SNMPv3 encryption protocol

    Methods
    -------
    autodetect()
        Try to determine the device type.

    """

    def __init__(
        self,
        hostname: str,
        snmp_version: str = "v3",
        snmp_port: int = 161,
        community: Optional[str] = None,
        user: str = "",
        auth_key: str = "",
        encrypt_key: str = "",
        auth_proto: str = "sha",
        encrypt_proto: str = "aes128",
    ) -> None:

        # Check that the SNMP version is matching predefined type or raise ValueError
        if snmp_version == "v1" or snmp_version == "v2c":
            if not community:
                raise ValueError("SNMP version v1/v2c community must be set.")
        elif snmp_version == "v3":
            if not user:
                raise ValueError("SNMP version v3 user and password must be set")
        else:
            raise ValueError("SNMP version must be set to 'v1', 'v2c' or 'v3'")

        # Check that the SNMPv3 auth & priv parameters match allowed types
        self._snmp_v3_authentication = {
            "sha": cmdgen.usmHMACSHAAuthProtocol,
            "md5": cmdgen.usmHMACMD5AuthProtocol,
        }
        self._snmp_v3_encryption = {
            "des": cmdgen.usmDESPrivProtocol,
            "3des": cmdgen.usm3DESEDEPrivProtocol,
            "aes128": cmdgen.usmAesCfb128Protocol,
            "aes192": cmdgen.usmAesCfb192Protocol,
            "aes256": cmdgen.usmAesCfb256Protocol,
        }
        if auth_proto not in self._snmp_v3_authentication.keys():
            raise ValueError(
                "SNMP V3 'auth_proto' argument must be one of the following: {}".format(
                    self._snmp_v3_authentication.keys()
                )
            )
        if encrypt_proto not in self._snmp_v3_encryption.keys():
            raise ValueError(
                "SNMP V3 'encrypt_proto' argument must be one of the following: {}".format(
                    self._snmp_v3_encryption.keys()
                )
            )

        self.hostname = hostname
        self.snmp_version = snmp_version
        self.snmp_port = snmp_port
        self.community = community
        self.user = user
        self.auth_key = auth_key
        self.encrypt_key = encrypt_key
        self.auth_proto = self._snmp_v3_authentication[auth_proto]
        self.encryp_proto = self._snmp_v3_encryption[encrypt_proto]
        self._response_cache: Dict[str, str] = {}

    def _get_snmpv3(self, oid: str) -> str:
        """
        Try to send an SNMP GET operation using SNMPv3 for the specified OID.

        Parameters
        ----------
        oid : str
            The SNMP OID that you want to get.

        Returns
        -------
        string : str
            The string as part of the value from the OID you are trying to retrieve.
        """
        snmp_target = (self.hostname, self.snmp_port)
        cmd_gen = cmdgen.CommandGenerator()

        (error_detected, error_status, error_index, snmp_data) = cmd_gen.getCmd(
            cmdgen.UsmUserData(
                self.user,
                self.auth_key,
                self.encrypt_key,
                authProtocol=self.auth_proto,
                privProtocol=self.encryp_proto,
            ),
            cmdgen.UdpTransportTarget(snmp_target, timeout=1.5, retries=2),
            oid,
            lookupNames=True,
            lookupValues=True,
        )

        if not error_detected and snmp_data[0][1]:
            return str(snmp_data[0][1])
        return ""

    def _get_snmpv2c(self, oid: str) -> str:
        """
        Try to send an SNMP GET operation using SNMPv2 for the specified OID.

        Parameters
        ----------
        oid : str
            The SNMP OID that you want to get.

        Returns
        -------
        string : str
            The string as part of the value from the OID you are trying to retrieve.
        """
        snmp_target = (self.hostname, self.snmp_port)
        cmd_gen = cmdgen.CommandGenerator()

        (error_detected, error_status, error_index, snmp_data) = cmd_gen.getCmd(
            cmdgen.CommunityData(self.community),
            cmdgen.UdpTransportTarget(snmp_target, timeout=1.5, retries=2),
            oid,
            lookupNames=True,
            lookupValues=True,
        )

        if not error_detected and snmp_data[0][1]:
            return str(snmp_data[0][1])
        return ""

    def _get_snmp(self, oid: str) -> str:
        """Wrapper for generic SNMP call."""
        if self.snmp_version in ["v1", "v2c"]:
            return self._get_snmpv2c(oid)
        else:
            return self._get_snmpv3(oid)

    def autodetect(self) -> Optional[str]:
        """
        Try to guess the device_type using SNMP GET based on the SNMP_MAPPER dict. The type which
        is returned is directly matching the name in *netmiko.ssh_dispatcher.CLASS_MAPPER_BASE*
        dict.

        Thus you can use this name to retrieve automatically the right ConnectionClass

        Returns
        -------
        potential_type : str
            The name of the device_type that must be running.
        """
        # Convert SNMP_MAPPER to a list and sort by priority
        snmp_mapper_orig = []
        for k, v in SNMP_MAPPER.items():
            snmp_mapper_orig.append({k: v})
        snmp_mapper_list = sorted(
            snmp_mapper_orig, key=lambda x: list(x.values())[0]["priority"]  # type: ignore
        )
        snmp_mapper_list.reverse()

        for entry in snmp_mapper_list:
            for device_type, v in entry.items():
                oid: str = v["oid"]  # type: ignore
                regex: Pattern = v["expr"]

                # Used cache data if we already queryied this OID
                if self._response_cache.get(oid):
                    snmp_response = self._response_cache.get(oid)
                else:
                    snmp_response = self._get_snmp(oid)
                    self._response_cache[oid] = snmp_response

                # See if we had a match
                if re.search(regex, snmp_response):
                    assert isinstance(device_type, str)
                    return device_type

        return None



from typing import Any, Sequence, TextIO, Union
from netmiko.cisco_base_connection import CiscoSSHConnection
from netmiko.no_enable import NoEnable
from netmiko.no_config import NoConfig


class ZyxelSSH(NoEnable, NoConfig, CiscoSSHConnection):
    def disable_paging(self, *args: Any, **kwargs: Any) -> str:
        """No paging on Zyxel"""
        return ""

    def send_config_set(
        self,
        config_commands: Union[str, Sequence[str], TextIO, None] = None,
        exit_config_mode: bool = False,
        enter_config_mode: bool = False,
        **kwargs: Any
    ) -> str:
        """No config mode on Zyxel"""
        return super().send_config_set(
            config_commands=config_commands,
            exit_config_mode=exit_config_mode,
            enter_config_mode=enter_config_mode,
            **kwargs
        )

    def session_preparation(self) -> None:
        super().session_preparation()
        # Zyxel switches output ansi codes
        self.ansi_escape_codes = True



from typing import Any, Optional, Union, Sequence
from typing import TYPE_CHECKING
import re
from netmiko.cisco_base_connection import CiscoSSHConnection
from netmiko.cisco_base_connection import CiscoFileTransfer

if TYPE_CHECKING:
    from netmiko.base_connection import BaseConnection


class AristaBase(CiscoSSHConnection):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        cmd = "terminal width 511"
        # Arista will echo immediately and then when the device really responds (like NX-OS)
        self.set_terminal_width(command=cmd, pattern=r"Width set to")
        self.disable_paging(cmd_verify=False, pattern=r"Pagination disabled")
        self.set_base_prompt()

    def enable(
        self,
        cmd: str = "enable",
        pattern: str = "ssword",
        enable_pattern: Optional[str] = r"\#",
        re_flags: int = re.IGNORECASE,
    ) -> str:
        return super().enable(
            cmd=cmd, pattern=pattern, enable_pattern=enable_pattern, re_flags=re_flags
        )

    def check_config_mode(
        self, check_string: str = ")#", pattern: str = r"[>\#]"
    ) -> bool:
        """
        Checks if the device is in configuration mode or not.

        Arista, unfortunately, does this:
        loc1-core01(s1)#

        Can also be (s2)
        """
        self.write_channel(self.RETURN)
        output = self.read_until_pattern(pattern=pattern)
        output = output.replace("(s1)", "")
        output = output.replace("(s2)", "")
        return check_string in output

    def config_mode(
        self,
        config_command: str = "configure terminal",
        pattern: str = "",
        re_flags: int = 0,
    ) -> str:
        """Force arista to read pattern all the way to prompt on the next line."""

        if not re_flags:
            re_flags = re.DOTALL
        check_string = re.escape(")#")

        if not pattern:
            pattern = re.escape(self.base_prompt[:16])
            pattern = f"{pattern}.*{check_string}"
        return super().config_mode(
            config_command=config_command, pattern=pattern, re_flags=re_flags
        )

    def _enter_shell(self) -> str:
        """Enter the Bourne Shell."""
        output = self._send_command_str("bash", expect_string=r"[\$#]")
        return output

    def _return_cli(self) -> str:
        """Return to the CLI."""
        output = self._send_command_str("exit", expect_string=r"[#>]")
        return output


class AristaSSH(AristaBase):
    pass


class AristaTelnet(AristaBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        default_enter = kwargs.get("default_enter")
        kwargs["default_enter"] = "\r\n" if default_enter is None else default_enter
        super().__init__(*args, **kwargs)


class AristaFileTransfer(CiscoFileTransfer):
    """Arista SCP File Transfer driver."""

    def __init__(
        self,
        ssh_conn: "BaseConnection",
        source_file: str,
        dest_file: str,
        file_system: Optional[str] = "/mnt/flash",
        direction: str = "put",
        **kwargs: Any,
    ) -> None:
        return super().__init__(
            ssh_conn=ssh_conn,
            source_file=source_file,
            dest_file=dest_file,
            file_system=file_system,
            direction=direction,
            **kwargs,
        )

    def remote_space_available(self, search_pattern: str = "") -> int:
        """Return space available on remote device."""
        return self._remote_space_available_unix(search_pattern=search_pattern)

    def check_file_exists(self, remote_cmd: str = "") -> bool:
        """Check if the dest_file already exists on the file system (return boolean)."""
        return self._check_file_exists_unix(remote_cmd=remote_cmd)

    def remote_file_size(
        self, remote_cmd: str = "", remote_file: Optional[str] = None
    ) -> int:
        """Get the file size of the remote file."""
        return self._remote_file_size_unix(
            remote_cmd=remote_cmd, remote_file=remote_file
        )

    def remote_md5(
        self, base_cmd: str = "verify /md5", remote_file: Optional[str] = None
    ) -> str:
        if remote_file is None:
            if self.direction == "put":
                remote_file = self.dest_file
            elif self.direction == "get":
                remote_file = self.source_file
        remote_md5_cmd = f"{base_cmd} file:{self.file_system}/{remote_file}"
        dest_md5 = self.ssh_ctl_chan._send_command_str(remote_md5_cmd, read_timeout=600)
        dest_md5 = self.process_md5(dest_md5)
        return dest_md5

    def enable_scp(self, cmd: Union[str, Sequence[str], None] = None) -> None:
        raise NotImplementedError

    def disable_scp(self, cmd: Union[str, Sequence[str], None] = None) -> None:
        raise NotImplementedError



from netmiko.cisco_base_connection import CiscoBaseConnection
from netmiko.no_enable import NoEnable
import time


class SmciSwitchSmisBase(NoEnable, CiscoBaseConnection):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self.config_mode()
        self.disable_paging(command="set cli pagination off")
        self.set_terminal_width(command="terminal width 511")
        self.exit_config_mode()
        # Clear the read buffer
        time.sleep(0.3 * self.global_delay_factor)
        self.clear_buffer()

    def check_enable_mode(self, check_string: str = "#") -> bool:
        """Check if in enable mode. Return boolean."""
        return super().check_enable_mode(check_string=check_string)

    def save_config(
        self,
        cmd: str = "write startup-config",
        confirm: bool = False,
        confirm_response: str = "",
    ) -> str:
        """Save config"""
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )


class SmciSwitchSmisSSH(SmciSwitchSmisBase):
    pass


class SmciSwitchSmisTelnet(SmciSwitchSmisBase):
    pass



import time
from socket import socket
from typing import Any

from netmiko.cisco_base_connection import CiscoBaseConnection
from telnetlib import IAC, DO, DONT, WILL, WONT, SB, SE, ECHO, SGA, NAWS, Telnet


class ZteZxrosBase(CiscoBaseConnection):
    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read(pattern=r"[>#]")
        self.set_base_prompt()
        self.disable_paging()
        # Clear the read buffer
        time.sleep(0.3 * self.global_delay_factor)
        self.clear_buffer()

    def check_config_mode(self, check_string: str = ")#", pattern: str = "#") -> bool:
        """
        Checks if the device is in configuration mode or not.
        """
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def save_config(
        self, cmd: str = "write", confirm: bool = False, confirm_response: str = ""
    ) -> str:
        """Saves Config Using Copy Run Start"""
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )


class ZteZxrosSSH(ZteZxrosBase):
    pass


class ZteZxrosTelnet(ZteZxrosBase):
    @staticmethod
    def _process_option(telnet_sock: socket, cmd: bytes, opt: bytes) -> None:
        """
        ZTE need manually reply DO ECHO to enable echo command.
        enable ECHO, SGA, set window size to [500, 50]
        """
        if cmd == WILL:
            if opt in [ECHO, SGA]:
                # reply DO ECHO / DO SGA
                telnet_sock.sendall(IAC + DO + opt)
            else:
                telnet_sock.sendall(IAC + DONT + opt)
        elif cmd == DO:
            if opt == NAWS:
                # negotiate about window size
                telnet_sock.sendall(IAC + WILL + opt)
                # Width:500, Height:50
                telnet_sock.sendall(IAC + SB + NAWS + b"\x01\xf4\x00\x32" + IAC + SE)
            else:
                telnet_sock.sendall(IAC + WONT + opt)

    def telnet_login(self, *args: Any, **kwargs: Any) -> str:
        # set callback function to handle telnet options.
        assert isinstance(self.remote_conn, Telnet)
        self.remote_conn.set_option_negotiation_callback(self._process_option)
        return super().telnet_login(*args, **kwargs)



"""Subclass specific to Cisco ASA."""
from typing import Any, Union, List, Dict, Optional
import re
import time
from netmiko.cisco_base_connection import CiscoSSHConnection, CiscoFileTransfer
from netmiko.exceptions import NetmikoAuthenticationException


class CiscoAsaSSH(CiscoSSHConnection):
    """Subclass specific to Cisco ASA."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs.setdefault("allow_auto_change", True)
        return super().__init__(*args, **kwargs)

    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""

        # Make sure the ASA is ready
        command = "show curpriv\n"
        self.write_channel(command)
        self.read_until_pattern(pattern=re.escape(command.strip()))

        # The 'enable' call requires the base_prompt to be set.
        self.set_base_prompt()
        if self.secret:
            self.enable()
        else:
            self.asa_login()
        self.disable_paging(command="terminal pager 0")

        if self.allow_auto_change:
            try:
                self.send_config_set("terminal width 511")
            except ValueError:
                # Don't fail for the terminal width
                pass
        else:
            # Disable cmd_verify if the terminal width can't be set
            self.global_cmd_verify = False

        self.set_base_prompt()

    def check_config_mode(
        self, check_string: str = ")#", pattern: str = r"[>\#]"
    ) -> bool:
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def enable(
        self,
        cmd: str = "enable",
        pattern: str = "ssword",
        enable_pattern: Optional[str] = r"\#",
        re_flags: int = re.IGNORECASE,
    ) -> str:
        return super().enable(
            cmd=cmd, pattern=pattern, enable_pattern=enable_pattern, re_flags=re_flags
        )

    def send_command_timing(
        self, *args: Any, **kwargs: Any
    ) -> Union[str, List[Any], Dict[str, Any]]:
        """
        If the ASA is in multi-context mode, then the base_prompt needs to be
        updated after each context change.
        """
        output = super().send_command_timing(*args, **kwargs)
        if len(args) >= 1:
            command_string = args[0]
        else:
            command_string = kwargs["command_string"]
        if "changeto" in command_string:
            self.set_base_prompt()
        return output

    def send_command(
        self, *args: Any, **kwargs: Any
    ) -> Union[str, List[Any], Dict[str, Any]]:
        """
        If the ASA is in multi-context mode, then the base_prompt needs to be
        updated after each context change.
        """
        if len(args) >= 1:
            command_string = args[0]
        else:
            command_string = kwargs["command_string"]

        # If changeto in command, look for '#' to determine command is done
        if "changeto" in command_string:
            if len(args) <= 1:
                expect_string = kwargs.get("expect_string", "#")
                kwargs["expect_string"] = expect_string
        output = super().send_command(*args, **kwargs)

        if "changeto" in command_string:
            self.set_base_prompt()

        return output

    def set_base_prompt(self, *args: Any, **kwargs: Any) -> str:
        """
        Cisco ASA in multi-context mode needs to have the base prompt updated
        (if you switch contexts i.e. 'changeto')

        This switch of ASA contexts can occur in configuration mode. If this
        happens the trailing '(config*' needs stripped off.
        """
        cur_base_prompt = super().set_base_prompt(*args, **kwargs)
        match = re.search(r"(.*)\(conf.*", cur_base_prompt)
        if match:
            # strip off (conf.* from base_prompt
            self.base_prompt = match.group(1)
            return self.base_prompt
        else:
            return cur_base_prompt

    def asa_login(self) -> None:
        """
        Handle ASA reaching privilege level 15 using login

        twb-dc-fw1> login
        Username: admin

        Raises NetmikoAuthenticationException, if we do not reach privilege
        level 15 after 10 loops.
        """
        delay_factor = self.select_delay_factor(0)

        i = 1
        max_attempts = 10
        self.write_channel("login" + self.RETURN)
        output = self.read_until_pattern(pattern=r"login")
        while i <= max_attempts:
            time.sleep(0.5 * delay_factor)
            output = self.read_channel()
            if "sername" in output:
                assert isinstance(self.username, str)
                self.write_channel(self.username + self.RETURN)
            elif "ssword" in output:
                assert isinstance(self.password, str)
                self.write_channel(self.password + self.RETURN)
            elif "#" in output:
                return
            else:
                self.write_channel("login" + self.RETURN)
            i += 1

        msg = "Unable to enter enable mode!"
        raise NetmikoAuthenticationException(msg)

    def save_config(
        self, cmd: str = "write mem", confirm: bool = False, confirm_response: str = ""
    ) -> str:
        """Saves Config"""
        return super().save_config(
            cmd=cmd, confirm=confirm, confirm_response=confirm_response
        )

    def normalize_linefeeds(self, a_string: str) -> str:
        """Cisco ASA needed that extra \r\n\r"""
        newline = re.compile("(\r\n\r|\r\r\r\n|\r\r\n|\r\n|\n\r)")
        a_string = newline.sub(self.RESPONSE_RETURN, a_string)
        if self.RESPONSE_RETURN == "\n":
            # Delete any remaining \r
            return re.sub("\r", "", a_string)
        else:
            return a_string


class CiscoAsaFileTransfer(CiscoFileTransfer):
    """Cisco ASA SCP File Transfer driver."""

    pass



#!/usr/bin/env python
import re
import pytest
from network_utilities import generate_ios_acl
from network_utilities import generate_cisco_nxos_acl  # noqa
from network_utilities import generate_cisco_asa_acl  # noqa
from network_utilities import generate_cisco_xr_acl  # noqa
from network_utilities import generate_arista_eos_acl  # noqa
from network_utilities import generate_juniper_junos_acl  # noqa


def remove_acl(net_connect, cmd, commit=False):
    """Ensure ACL is removed."""
    net_connect.send_config_set(cmd)
    if commit:
        net_connect.commit()
        net_connect.exit_config_mode()


def test_large_acl(net_connect, commands, expected_responses, acl_entries=100):
    """Test configuring a large ACL."""
    if commands.get("config_long_acl"):
        acl_config = commands.get("config_long_acl")
        base_cmd = acl_config["base_cmd"]
        verify_cmd = acl_config["verify_cmd"]
        delete_cmd = acl_config.get("delete_cmd")
        offset = acl_config["offset"]
    else:
        pytest.skip("Platform not supported for ACL test")

    platform = net_connect.device_type
    support_commit = commands.get("support_commit")

    if "juniper_junos" in platform or "cisco_asa" in platform:
        cmd = delete_cmd
    else:
        cmd = f"no {base_cmd}"
    remove_acl(net_connect, cmd=cmd, commit=support_commit)

    # Generate the ACL
    platform = net_connect.device_type
    if "cisco_ios" in net_connect.device_type or "cisco_xe" in net_connect.device_type:
        cfg_lines = generate_ios_acl()
    else:
        func_name = f"generate_{platform}_acl"
        acl_func = globals()[func_name]
        cfg_lines = acl_func()

    # Send ACL to remote devices
    result = net_connect.send_config_set(cfg_lines)
    if support_commit:
        net_connect.commit()
        net_connect.exit_config_mode()

    # send_config_set should return same num lines + offset lines for entering/exiting cfg-mode
    # NX-OS is will have more than one newline (per line)
    result_list = re.split(r"\n+", result)
    assert len(result_list) == len(cfg_lines) + offset

    # Check that length of lines in show of the acl matches lines configured
    verify = net_connect.send_command(verify_cmd, read_timeout=90)
    verify_list = re.split(r"\n+", verify.strip())

    # IOS-XR potentially has a timestamp on the show command
    if "UTC" in verify_list[0]:
        verify_list.pop(0)
    if "juniper_junos" in platform:
        offset = 6
        assert len(verify_list) - offset == len(cfg_lines)
    elif "cisco_asa" in platform:
        offset = 1
        assert len(verify_list) - offset == len(cfg_lines)
    else:
        assert len(verify_list) == len(cfg_lines)

    if "juniper_junos" in platform or "cisco_asa" in platform:
        cmd = delete_cmd
    else:
        cmd = f"no {base_cmd}"
    remove_acl(net_connect, cmd=cmd, commit=support_commit)
    net_connect.disconnect()



from ipaddress import ip_address


def generate_acl(
    acl_name="netmiko_test_large_acl",
    entries=100,
    base_cmd="ip access-list extended",
    base_addr="192.168.0.0",
):
    cmd = f"{base_cmd} {acl_name}"
    acl = [cmd]
    for i in range(1, entries + 1):
        addr = ip_address(base_addr)
        cmd = f"permit ip host {addr + i} any"
        acl.append(cmd)
    return acl


def generate_hp_procurve_acl(
    acl_name="netmiko_test_large_acl",
    entries=100,
    base_cmd="ip access-list extended",
    base_addr="192.168.0.0",
):
    """
    Faked an ACL as did not see easy way to create ACL on my HPE device.

    Just remove and add NTP server so 100 lines of cfg changes.
    """

    base_acl = [
        "no sntp server 128.118.25.3",
        "sntp server 128.118.25.3",
    ]
    acl = []
    for _ in range(50):
        acl += base_acl
    return acl


def generate_ios_acl(
    acl_name="netmiko_test_large_acl",
    entries=100,
    base_cmd="ip access-list extended",
    base_addr="192.168.0.0",
):
    return generate_acl(
        acl_name=acl_name, entries=entries, base_cmd=base_cmd, base_addr=base_addr
    )


def generate_arista_eos_acl(
    acl_name="netmiko_test_large_acl",
    entries=100,
    base_cmd="ip access-list",
    base_addr="192.168.0.0",
):
    return generate_acl(
        acl_name=acl_name, entries=entries, base_cmd=base_cmd, base_addr=base_addr
    )


generate_cisco_ios_acl = generate_ios_acl
generate_cisco_xe_acl = generate_ios_acl


def generate_nxos_acl(
    acl_name="netmiko_test_large_acl",
    entries=100,
    base_cmd="ip access-list",
    base_addr="192.168.0.0",
):
    return generate_acl(
        acl_name=acl_name, entries=entries, base_cmd=base_cmd, base_addr=base_addr
    )


generate_cisco_nxos_acl = generate_nxos_acl


def generate_cisco_xr_acl(
    acl_name="netmiko_test_large_acl",
    entries=100,
    base_cmd="ipv4 access-list",
    base_addr="192.168.0.0",
):

    # Add base ACL command
    cmd = f"{base_cmd} {acl_name}"
    acl = [cmd]
    for i in range(1, entries + 1):
        addr = ip_address(base_addr)
        cmd = f"permit ipv4 host {addr + i} any"
        acl.append(cmd)
    return acl


def generate_juniper_junos_acl(
    acl_name="netmiko_test_large_acl",
    entries=100,
    base_cmd="set firewall family inet filter",
    base_addr="192.168.0.0",
):
    acl = []
    for i in range(1, entries + 1):
        addr = ip_address(base_addr)
        cmd = f"{base_cmd} {acl_name} term 10 from address {addr + i}"
        acl.append(cmd)
    return acl


def generate_cisco_asa_acl(
    acl_name="netmiko_test_large_acl",
    entries=100,
    base_cmd=None,
    base_addr="192.168.0.0",
):
    acl = []
    for i in range(1, entries + 1):
        addr = ip_address(base_addr)
        cmd = f"access-list {acl_name} extended permit ip host {addr + i} any"
        acl.append(cmd)
    return acl


if __name__ == "__main__":
    # Test code
    acl = generate_ios_acl(entries=10)
    ios_ref_acl = [
        "ip access-list extended netmiko_test_large_acl",
        "permit ip host 192.168.0.1 any",
        "permit ip host 192.168.0.2 any",
        "permit ip host 192.168.0.3 any",
        "permit ip host 192.168.0.4 any",
        "permit ip host 192.168.0.5 any",
        "permit ip host 192.168.0.6 any",
        "permit ip host 192.168.0.7 any",
        "permit ip host 192.168.0.8 any",
        "permit ip host 192.168.0.9 any",
        "permit ip host 192.168.0.10 any",
    ]

    assert acl == ios_ref_acl

    acl = generate_nxos_acl(entries=10)
    nxos_ref_acl = [
        "ip access-list netmiko_test_large_acl",
        "permit ip host 192.168.0.1 any",
        "permit ip host 192.168.0.2 any",
        "permit ip host 192.168.0.3 any",
        "permit ip host 192.168.0.4 any",
        "permit ip host 192.168.0.5 any",
        "permit ip host 192.168.0.6 any",
        "permit ip host 192.168.0.7 any",
        "permit ip host 192.168.0.8 any",
        "permit ip host 192.168.0.9 any",
        "permit ip host 192.168.0.10 any",
    ]
    assert acl == nxos_ref_acl

    acl = generate_cisco_xr_acl(entries=10)
    xr_ref_acl = [
        "ipv4 access-list netmiko_test_large_acl",
        "permit ipv4 host 192.168.0.1 any",
        "permit ipv4 host 192.168.0.2 any",
        "permit ipv4 host 192.168.0.3 any",
        "permit ipv4 host 192.168.0.4 any",
        "permit ipv4 host 192.168.0.5 any",
        "permit ipv4 host 192.168.0.6 any",
        "permit ipv4 host 192.168.0.7 any",
        "permit ipv4 host 192.168.0.8 any",
        "permit ipv4 host 192.168.0.9 any",
        "permit ipv4 host 192.168.0.10 any",
    ]
    assert acl == xr_ref_acl

    acl = generate_juniper_junos_acl(entries=10)
    ref_acl = [
        "set firewall family inet filter netmiko_test_large_acl term 10 from address "
        "192.168.0.1",
        "set firewall family inet filter netmiko_test_large_acl term 10 from address "
        "192.168.0.2",
        "set firewall family inet filter netmiko_test_large_acl term 10 from address "
        "192.168.0.3",
        "set firewall family inet filter netmiko_test_large_acl term 10 from address "
        "192.168.0.4",
        "set firewall family inet filter netmiko_test_large_acl term 10 from address "
        "192.168.0.5",
        "set firewall family inet filter netmiko_test_large_acl term 10 from address "
        "192.168.0.6",
        "set firewall family inet filter netmiko_test_large_acl term 10 from address "
        "192.168.0.7",
        "set firewall family inet filter netmiko_test_large_acl term 10 from address "
        "192.168.0.8",
        "set firewall family inet filter netmiko_test_large_acl term 10 from address "
        "192.168.0.9",
        "set firewall family inet filter netmiko_test_large_acl term 10 from address "
        "192.168.0.10",
    ]
    assert acl == ref_acl

    acl = generate_cisco_asa_acl(entries=10)
    ref_acl = [
        "access-list netmiko_test_large_acl extended permit ip host 192.168.0.1 any",
        "access-list netmiko_test_large_acl extended permit ip host 192.168.0.2 any",
        "access-list netmiko_test_large_acl extended permit ip host 192.168.0.3 any",
        "access-list netmiko_test_large_acl extended permit ip host 192.168.0.4 any",
        "access-list netmiko_test_large_acl extended permit ip host 192.168.0.5 any",
        "access-list netmiko_test_large_acl extended permit ip host 192.168.0.6 any",
        "access-list netmiko_test_large_acl extended permit ip host 192.168.0.7 any",
        "access-list netmiko_test_large_acl extended permit ip host 192.168.0.8 any",
        "access-list netmiko_test_large_acl extended permit ip host 192.168.0.9 any",
        "access-list netmiko_test_large_acl extended permit ip host 192.168.0.10 any",
    ]
    assert acl == ref_acl



from typing import Optional, Union, Sequence, TextIO, Any
import time
import warnings
import re
from netmiko.no_enable import NoEnable
from netmiko.base_connection import DELAY_FACTOR_DEPR_SIMPLE_MSG
from netmiko.cisco_base_connection import CiscoSSHConnection


class VyOSSSH(NoEnable, CiscoSSHConnection):
    """Implement methods for interacting with VyOS network devices."""

    def session_preparation(self) -> None:
        """Prepare the session after the connection has been established."""
        self._test_channel_read()
        self.set_base_prompt()
        self.set_terminal_width(command="set terminal width 512", pattern="terminal")
        self.disable_paging(command="set terminal length 0")
        # Clear the read buffer
        time.sleep(0.3 * self.global_delay_factor)
        self.clear_buffer()

    def check_config_mode(self, check_string: str = "#", pattern: str = "") -> bool:
        """Checks if the device is in configuration mode"""
        return super().check_config_mode(check_string=check_string, pattern=pattern)

    def config_mode(
        self,
        config_command: str = "configure",
        pattern: str = r"\[edit\]",
        re_flags: int = 0,
    ) -> str:
        return super().config_mode(
            config_command=config_command, pattern=pattern, re_flags=re_flags
        )

    def exit_config_mode(
        self, exit_config: str = "exit", pattern: str = r"exit"
    ) -> str:
        """Exit configuration mode"""
        output = ""
        if self.check_config_mode():
            output = self._send_command_timing_str(
                exit_config, strip_prompt=False, strip_command=False
            )
            if "Cannot exit: configuration modified" in output:
                output += self._send_command_timing_str(
                    "exit discard", strip_prompt=False, strip_command=False
                )
            if self.check_config_mode():
                raise ValueError("Failed to exit configuration mode")
        return output

    def commit(
        self,
        comment: str = "",
        read_timeout: float = 120.0,
        delay_factor: Optional[float] = None,
    ) -> str:
        """
        Commit the candidate configuration.

        Commit the entered configuration. Raise an error and return the failure
        if the commit fails.

        default:
           command_string = commit
        comment:
           command_string = commit comment <comment>

        delay_factor: Deprecated in Netmiko 4.x. Will be eliminated in Netmiko 5.

        """

        if delay_factor is not None:
            warnings.warn(DELAY_FACTOR_DEPR_SIMPLE_MSG, DeprecationWarning)

        error_marker = ["Failed to generate committed config", "Commit failed"]
        command_string = "commit"

        if comment:
            command_string += f' comment "{comment}"'

        output = self.config_mode()
        output += self._send_command_str(
            command_string,
            strip_prompt=False,
            strip_command=False,
            read_timeout=read_timeout,
        )

        if any(x in output for x in error_marker):
            raise ValueError(f"Commit failed with following errors:\n\n{output}")
        return output

    def set_base_prompt(
        self,
        pri_prompt_terminator: str = "$",
        alt_prompt_terminator: str = "#",
        delay_factor: float = 1.0,
        pattern: Optional[str] = None,
    ) -> str:
        """Sets self.base_prompt: used as delimiter for stripping of trailing prompt in output."""

        # VyOS can have a third terminator; switch to a pattern solution
        if pattern is None:
            pri_term = re.escape(pri_prompt_terminator)
            alt_term = re.escape(alt_prompt_terminator)
            third_terminator = re.escape(">")
            pattern = rf"({pri_term}|{alt_term}|{third_terminator})"

        prompt = super().set_base_prompt(
            pri_prompt_terminator=pri_prompt_terminator,
            alt_prompt_terminator=alt_prompt_terminator,
            delay_factor=delay_factor,
            pattern=pattern,
        )
        # Set prompt to user@hostname (remove two additional characters)
        self.base_prompt = prompt[:-2].strip()
        return self.base_prompt

    def send_config_set(
        self,
        config_commands: Union[str, Sequence[str], TextIO, None] = None,
        exit_config_mode: bool = False,
        **kwargs: Any,
    ) -> str:
        """Remain in configuration mode."""
        return super().send_config_set(
            config_commands=config_commands, exit_config_mode=exit_config_mode, **kwargs
        )

    def save_config(
        self,
        cmd: str = "copy running-config startup-config",
        confirm: bool = False,
        confirm_response: str = "",
    ) -> str:
        """Not Implemented"""
        raise NotImplementedError



from typing import Optional
from netmiko.no_enable import NoEnable
from netmiko.no_config import NoConfig
from netmiko.cisco_base_connection import CiscoSSHConnection


class AccedianSSH(NoEnable, NoConfig, CiscoSSHConnection):
    def session_preparation(self) -> None:
        self._test_channel_read(pattern=r"[:#]")
        self.set_base_prompt()

    def set_base_prompt(
        self,
        pri_prompt_terminator: str = ":",
        alt_prompt_terminator: str = "#",
        delay_factor: float = 2.0,
        pattern: Optional[str] = None,
    ) -> str:
        """Sets self.base_prompt: used as delimiter for stripping of trailing prompt in output."""
        super().set_base_prompt(
            pri_prompt_terminator=pri_prompt_terminator,
            alt_prompt_terminator=alt_prompt_terminator,
            delay_factor=delay_factor,
            pattern=pattern,
        )
        return self.base_prompt

    def save_config(
        self, cmd: str = "", confirm: bool = False, confirm_response: str = ""
    ) -> str:
        """Not Implemented"""
        raise NotImplementedError



from datetime import datetime


def execute_cmd(conn, cmd="show tech-support", read_timeout=None, last_read=2.0):
    start_time = datetime.now()
    if read_timeout is None:
        output = conn.send_command_timing(cmd, last_read=last_read, strip_prompt=False)
    else:
        output = conn.send_command_timing(
            cmd, read_timeout=read_timeout, last_read=last_read, strip_prompt=False
        )
    end_time = datetime.now()
    exec_time = end_time - start_time
    return (output, exec_time)


def test_read_show_tech(net_connect_newconn):

    read_timeout = 0
    output, exec_time = execute_cmd(
        net_connect_newconn, read_timeout=read_timeout, last_read=4.0
    )
    assert "show interface" in output
    assert "cisco3#" in output
    assert exec_time.total_seconds() > 10
    net_connect_newconn.disconnect()


def test_read_md5(net_connect_newconn):

    cmd = "verify /sha512 flash:/c1100-universalk9_ias.16.12.03.SPA.bin\n"
    # cmd = "verify /md5 flash:/c1100-universalk9_ias.16.12.03.SPA.bin\n"
    output, exec_time = execute_cmd(net_connect_newconn, cmd=cmd)
    assert "Done!" in output
    assert "cisco3#" in output
    assert exec_time.total_seconds() > 10
    net_connect_newconn.disconnect()


def test_read_ping(net_connect_newconn):

    cmd = "ping 8.8.8.8 repeat 10000\n"
    output, exec_time = execute_cmd(net_connect_newconn, cmd=cmd)
    assert "Success rate is" in output
    assert "cisco3#" in output
    assert exec_time.total_seconds() > 10
    net_connect_newconn.disconnect()


def test_read_ping_no_response(net_connect_newconn):

    # Nothing exists at that IP
    cmd = "ping 10.220.88.209 repeat 20\n"
    output, exec_time = execute_cmd(
        net_connect_newconn,
        cmd=cmd,
    )
    assert "Success rate is" in output
    assert "cisco3#" in output
    assert exec_time.total_seconds() > 10
    net_connect_newconn.disconnect()


def test_read_traceroute_no_response(net_connect_newconn):

    # Will finish in about 90 seconds
    cmd = "traceroute 8.8.8.8 probe 1"
    output, exec_time = execute_cmd(
        net_connect_newconn,
        cmd=cmd,
        last_read=4.0,
        #    read_timeout=300,
    )
    assert "cisco3#" in output
    assert output.count("*") == 30
    assert exec_time.total_seconds() > 10
    net_connect_newconn.disconnect()


def test_read_traceroute_no_response_full(net_connect_newconn):

    # Will finish in about 4.5 minutes
    cmd = "traceroute 8.8.8.8"
    output, exec_time = execute_cmd(
        net_connect_newconn,
        cmd=cmd,
        last_read=4.0,
        read_timeout=6 * 60,  # allow 6-minutes
    )
    assert "cisco3#" in output
    assert output.count("*") == 90
    assert exec_time.total_seconds() > 100
    net_connect_newconn.disconnect()


def test_read_traceroute(net_connect_newconn):

    cmd = "traceroute 10.220.88.37"
    output, exec_time = execute_cmd(
        net_connect_newconn,
        cmd=cmd,
        last_read=4.0,
    )
    assert "cisco3#" in output
    assert exec_time.total_seconds() > 5
    net_connect_newconn.disconnect()



from typing import Any
from paramiko import SSHClient


class SSHClient_noauth(SSHClient):
    """Set noauth when manually handling SSH authentication."""

    def _auth(self, username: str, *args: Any) -> None:
        transport = self.get_transport()
        assert transport is not None
        transport.auth_none(username)
        return
